#include <stdio.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <signal.h>
#include "global.h"
#include "config.h"
#include "mbuf.h"
#include "iface.h"
#include "sems.h"
#include "ax25.h"
#include "rpc.h"
#include "memory.h"
#include "usock.h"

extern struct mbuf *slip_decode();
struct mbuf **Hopper;
struct memory Mainmem;
extern struct memory *Mem;
int Intoff = 0, *Selectcnt;
void iostop();
struct usock *Usock = NULLUSOCK; /* Dummy to please some functions in proc.c */

main(argc,argv)
int argc;
char **argv;
{
    int fd, parent, type;
    char c;
    struct iface *iface;
    struct mbuf *bp,*nbp;
    struct phdr *phdr;
    int nc;
    char buf[256];
    register char *cp;

    if (argc < 4) {
	fprintf(stderr,"Usage: asy_rx fd iface type\n");
	exit(1);
    }
    parent = getppid();
    fd = atoi(argv[1]);
    iface = (struct iface *) atoi(argv[2]);
    type = atoi(argv[3]);
    init();
    nc = 0;
    for(;;){
	if (nc-- == 0) {
	    cp = &buf[0];
	    nc = read(fd,cp,sizeof(buf));
	    if (nc < 1)
		iostop();
	    nc--;
	}
	if((bp = slip_decode(0,*cp++)) == NULLBUF)
	    continue;	/* More to come */

	if((nbp = pushdown(bp,sizeof(struct phdr))) == NULLBUF){
	    free_p(bp);
	    continue;
	}
	phdr = (struct phdr *)nbp->data;
	phdr->iface = iface;
	phdr->type = type;
	enqueue(Hopper,nbp);
	/* Instead of including all of rpcall(), we do the following: */
	bp = alloc_mbuf(sizeof(int));
	if(bp == NULLBUF)
	    printf("asy_rx: malloc failed!\n");
	bp->cnt = sizeof(int);
	*(int *)bp->data = RPC_STOPWAIT;
	rpcsend(bp,1);
	if (kill(parent, 0) == -1 && errno == ESRCH) /* parent exited? */
	    iostop();
    }
}

init()
{
    int32 addr;
    struct mrec *mrec;

    signal(SIGINT, iostop);
    signal(SIGTERM, iostop);
    signal(SIGHUP, iostop);
    signal(SIGQUIT, iostop);

    if ((Mainmem.semid = semget(KEY_CNTL, 0, 0)) == -1)
	perror("semget server");
    grabcore(&Mainmem, 0, 0, KEY_BUFFER);
    mrec = (struct mrec *) Mainmem.shmbase;
    /* Relocate memory at the right address if necessary */
    if (Mainmem.shmbase != (char *)mrec->data[0]) {
	addr = mrec->data[0];
	shmdt(Mainmem.shmbase);
	grabcore(&Mainmem, 0, addr, KEY_BUFFER);
	mrec = (struct mrec *) Mainmem.shmbase;
    }
    Mainmem.block = 1;
    Mainmem.queue = (struct mbuf **) mrec->data[2];
    Hopper = (struct mbuf **) mrec->data[3];
    Selectcnt = (int *) mrec->data[4];
}

void
iostop()
{
    if (Intoff)
	restore(1);
    if (shmdt(Mainmem.shmbase) == -1)
	perror("shmdt");
    exit(0);
}

/* Restore the "interrupts" */
void restore (mask)
char mask;
{
    struct sembuf sembuf;
    if (mask) {
	sembuf.sem_op = INC_SEM;
	sembuf.sem_num = SEM_MEMORY;
	sembuf.sem_flg = 0;
	semop(Mem->semid, &sembuf, 1);
	Intoff = 0;
    }
}
coreleft()
{
    return 0;
}
