#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "timer.h"
#include "iface.h"
#include "ax25.h"
#include "lapb.h"
#include "cmdparse.h"

extern char Notval[];
extern int Digipeat;
extern int16 T3init,N2,Maxframe,Paclen,Pthresh,Axwindow,Axversion;
extern int32 Axirtt;
extern int Axi_sock;

static int domycall(),dodigipeat(),doaxirtt(),doaxstat(),dot3(),domaxframe(),
	doaxwindow(),dopaclen(),don2(),doaxreset(),dopthresh(),doaxroute(),
	doversion(),doaxkick();

char *Ax25states[] = {
	"Disconnected",
	"Listening",
	"Conn pending",
	"Disc pending",
	"Connected",
	"Recovery",
	"Frame Reject",
};

/* Ascii explanations for the disconnect reasons listed in lapb.h under
 * "reason" in ax25_cb
 */
char *Axreasons[] = {
	"Normal",
	"DM received",
	"Timeout"
};

static struct cmds axcmds[] = {
	"digipeat",	dodigipeat,	0, 0, NULLCHAR,
	"irtt",		doaxirtt,	0, 0, NULLCHAR,
	"kick",		doaxkick,	0, 2, "ax25 kick <axcb>",
	"maxframe",	domaxframe,	0, 0, NULLCHAR,
	"mycall",	domycall,	0, 0, NULLCHAR,
	"paclen",	dopaclen,	0, 0, NULLCHAR,
	"pthresh",	dopthresh,	0, 0, NULLCHAR,
	"reset",	doaxreset,	0, 2, "ax25 reset <axcb>",
	"retry",	don2,		0, 0, NULLCHAR,
	"route",	doaxroute,	0, 0, NULLCHAR,
	"status",	doaxstat,	0, 0, NULLCHAR,
	"t3",		dot3,		0, 0, NULLCHAR,
	"version",	doversion,	0, 0, NULLCHAR,
	"window",	doaxwindow,	0, 0, NULLCHAR,
	NULLCHAR,	NULLFP,		0, 0, "ax25 subcommands: digipeat irtt kick maxframe mycall paclen pthresh reset\n\t\t  retry route status t3 version window",
};
/* Multiplexer for top-level ax25 command */
doax25(argc,argv,envp)
int argc;
char *argv[];
void *envp;
{
	return subcmd(axcmds,argc,argv,envp);
}

static
doaxreset(argc,argv)
int argc;
char *argv[];
{
	struct ax25_cb *axp;

	axp = (struct ax25_cb *)ltop(htol(argv[1]));
	if(!ax25val(axp)){
		printf(Notval);
		return 1;
	}
	reset_ax25(axp);
	return 0;
}

/* Display AX.25 link level control blocks */
static
doaxstat(argc,argv)
int argc;
char *argv[];
{
	register int i;
	register struct ax25_cb *axp;
	char tmp[10];

	if(argc < 2){
		printf("    &AXB Snd-Q   Rcv-Q   Remote    State\n");
		for(i=0;i<NHASH;i++){
			for(axp = Ax25_cb[i];axp != NULLAX25; axp = axp->next){
				pax25(tmp,&axp->remote);
				printf("%8lx %-8d%-8d%-10s%s\n",
					ptol(axp),
					len_q(axp->txq),len_mbuf(axp->rxq),
					tmp,Ax25states[axp->state]);
			}
		}
		return 0;
	}
	axp = (struct ax25_cb *)ltop(htol(argv[1]));
	if(!ax25val(axp)){
		printf(Notval);
		return 1;
	}
	st_ax25(axp);
	return 0;
}
/* Dump one control block */
void
st_ax25(axp)
register struct ax25_cb *axp;
{
	char tmp[10];

	if(axp == NULLAX25)
		return;
	printf("    &AXB Remote   RB V(S) V(R) Unack P Retry State\n");
	pax25(tmp,&axp->remote);
	printf("%8lx %-9s",ptol(axp),tmp);
	putchar(axp->flags.rejsent ? 'R' : ' ');
	putchar(axp->flags.remotebusy ? 'B' : ' ');
	printf(" %4d %4d",axp->vs,axp->vr);
	printf(" %02u/%02u %u",axp->unack,axp->maxframe,axp->proto);
	printf(" %02u/%02u",axp->retries,axp->n2);
	printf(" %s\n",Ax25states[axp->state]);

	printf("SRT = %lu ",axp->srt * MSPTICK);
	printf("T1: ");
	if(run_timer(&axp->t1))
		printf("%lu",read_timer(&axp->t1) * MSPTICK);
	else
		printf("stop");
	printf("/%lu ms; ",dur_timer(&axp->t1) * MSPTICK);

	printf("T3: ");
	if(run_timer(&axp->t3))
		printf("%lu",read_timer(&axp->t3) * MSPTICK);
	else
		printf("stop");
	printf("/%lu ms\n",dur_timer(&axp->t3) * MSPTICK);

}

/* Display or change our AX.25 address */
static
domycall(argc,argv)
int argc;
char *argv[];
{
	char buf[15];

	if(argc < 2){
		pax25(buf,&Mycall);
		printf("%s\n",buf);
		return 0;
	}
	if(setcall(&Mycall,argv[1]) == -1)
		return -1;
	Mycall.ssid |= E;
	return 0;
}

/* Control AX.25 digipeating */
static
dodigipeat(argc,argv)
int argc;
char *argv[];
{
	if(argc == 1) {
		printf("digipeat %s\n",Digipeat ? "on" : "off");
	} else {
		if(strcmp(argv[1],"on") == 0)
			Digipeat = 1;
		else
			Digipeat = 0;
	}
	return 0;
}
static
doversion(argc,argv)
int argc;
char *argv[];
{
	if(argc == 1){
		printf("AX25 version %s\n",Axversion == V1 ? "V1" : "V2");
		return 0;
	}
	if(strchr(argv[1],'1') != NULLCHAR)
		Axversion = V1;
	else if(strchr(argv[1],'2') != NULLCHAR)
		Axversion = V2;
	else
		return -1;
	return 0;
}

static
doaxirtt(argc,argv)
int argc;
char *argv[];
{
	if(argc == 1){
		printf("AX25 initial round trip time: %lu ms\n",Axirtt*MSPTICK);
	} else {
		Axirtt = atol(argv[1]) / MSPTICK;
	}
	return 0;
}

/* Set idle timer */
static
dot3(argc,argv)
int argc;
char *argv[];
{
	if(argc == 1) {
		printf("T3 %lu ms\n",(long)T3init * MSPTICK);
	} else {
		T3init = atol(argv[1]) / MSPTICK;
	}
	return 0;
}

/* Set retry limit count */
static
don2(argc,argv)
int argc;
char *argv[];
{
	if(argc == 1) {
		printf("Retry %u\n",N2);
	} else {
		N2 = atoi(argv[1]);
	}
	return 0;
}
/* Force a retransmission */
static
doaxkick(argc,argv)
int argc;
char *argv[];
{
	struct ax25_cb *axp;

	axp = (struct ax25_cb *)ltop(htol(argv[1]));
	if(!ax25val(axp)){
		printf(Notval);
		return 1;
	}
	kick_ax25(axp);
	return 0;
}
/* Set maximum number of frames that will be allowed in flight */
static
domaxframe(argc,argv)
int argc;
char *argv[];
{
	if(argc == 1) {
		printf("Maxframe %u\n",Maxframe);
	} else {
		Maxframe = atoi(argv[1]);
	}
	return 0;
}

/* Set maximum length of I-frame data field */
static
dopaclen(argc,argv)
int argc;
char *argv[];
{
	if(argc == 1) {
		printf("Paclen %u\n",Paclen);
	} else {
		Paclen = atoi(argv[1]);
	}
	return 0;
}
/* Set size of I-frame above which polls will be sent after a timeout */
static
dopthresh(argc,argv)
int argc;
char *argv[];
{
	if(argc == 1) {
		printf("Pthresh %u bytes\n",Pthresh);
	} else {
		Pthresh = atoi(argv[1]);
	}
	return 0;
}

/* Set high water mark on receive queue that triggers RNR */
static
doaxwindow(argc,argv)
int argc;
char *argv[];
{
	if(argc == 1) {
		printf("Axwindow %u\n",Axwindow);
	} else {
		Axwindow = atoi(argv[1]);
	}
	return 0;
}
/* End of ax25 subcommands */

/* Display and modify AX.25 routing table */
static int
doaxroute(argc,argv)
int argc;
char *argv[];
{
	char buf[30];
	int i,j,ndigis;
	register struct ax_route *axr;
	struct ax25_addr target,digis[MAXDIGIS];

	if(argc < 2){
		printf("Target    Type   Digipeaters\n");
		for(i=0;i<NAXROUTE;i++){
			for(axr = Ax_routes[i];axr != NULLAXR;axr = axr->next){
				pax25(buf,&axr->target);
				printf("%-10s%-6s",buf,
				 axr->type == AX_LOCAL ? "Local":"Auto");
				for(j=0;j<axr->ndigis;j++){
					pax25(buf,&axr->digis[j]);
					printf(" %s",buf);
				}
				printf("\n");
			}
		}
		return 0;
	}
	if(argc < 3){
		printf("Usage: ax25 route add <target> [digis...]\n");
		printf("       ax25 route drop <target>\n");
		return 1;
	}
	if(setcall(&target,argv[2]) == -1){
		printf("Bad target %s\n",argv[2]);
		return 1;
	}
	switch(argv[1][0]){
	case 'a':	/* Add route */
		if(argc < 3){
			printf("Usage: ax25 route add <target> [digis...]\n");
			return 1;
		}
		ndigis = argc - 3;
		for(i=0;i<ndigis;i++){
			if(setcall(&digis[i],argv[i+3]) == -1){
				printf("Bad digipeater %s\n",argv[i+3]);
				return 1;
			}
		}
		if(ax_add(&target,AX_LOCAL,&digis[0],ndigis) == NULLAXR){
			printf("Failed\n");
			return 1;
		}
		break;
	case 'd':	/* Drop route */
		if(ax_drop(&target) == -1){
			printf("Not in table\n");
			return 1;
		}
		break;
	default:
		printf("Unknown command %s\n",argv[1]);
		return 1;
	}
	return 0;
}
