#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include "global.h"
#include "config.h"
#include "sems.h"
#include "mbuf.h"
#include "memory.h"
#include "ax25.h"
#include "rpc.h"
#include "timer.h"
#include "usock.h"

char *shmalloc();
struct memory Mainmem, Ownmem;
extern struct memory *Mem;
int Usockid, Intoff=0;
key_t Mykey;
int *Selectcnt;

int
shminit()
{
    int32 addr;
    void iostop();
    struct mrec *mrec;
    extern int Initok;
    extern struct ax25_addr Mycall;
    extern int32 Ip_addr;
    struct settings *setp;
    Initok = 1;
    if ((Mainmem.semid = semget(KEY_CNTL, 0, 0)) == -1)
	return -1;

    signal(SIGINT, iostop);
    signal(SIGTERM, iostop);
    signal(SIGHUP, iostop);
    signal(SIGQUIT, iostop);

    grabcore(&Mainmem, 0, 0, KEY_BUFFER);
    mrec = (struct mrec *) Mainmem.shmbase;
    /* Relocate memory at the right address if necessary */
    if (Mainmem.shmbase != (char *)mrec->data[0]) {
	addr = mrec->data[0];
	shmdt(Mainmem.shmbase);
	grabcore(&Mainmem, 0, addr, KEY_BUFFER);
	mrec = (struct mrec *) Mainmem.shmbase;
    }
    Mainmem.queue = (struct mbuf **) mrec->data[2];
    Selectcnt = (int *) mrec->data[4];
    sockinit();
    if (Usock != (struct usock *)mrec->data[1]) {
	addr = mrec->data[1];
	shmdt(Usock);
	Usock = (struct usock *)shmat(Usockid, addr, 0);
    }
    setp = (struct settings *) rpcall(RPC_GETSET);
    Ip_addr = setp->Ip_addr;
    strcpy(&Mycall,&setp->Mycall,sizeof(struct ax25_addr));
    shfree(setp);

#ifdef notdef
    Mykey = ftok("/", KEY_CNTL);
    Ownmem.block = Mainmem.block = 1; /* malloc is not used in Ownmem though */
    grabcore(&Ownmem,0,rpcall(RPC_INIT,Mykey),Mykey);
    mrec = (struct mrec *) Ownmem.shmbase;
    Ownmem.queue = (struct mbuf **) mrec->data[2];
    printf("Ownmem at %d and queue at %d\n",Ownmem.base,Ownmem.queue);
    Mem = &Mainmem;	/* default memory is always Mainmem */
#endif
    return 0;
}

void
iostop()
{
    if (Intoff)
	restore(1);
    freesock(getpid());
    if (shmdt(Mainmem.shmbase) == -1)
	perror("shmdt");
#ifdef notdef
    if (shmdt(Ownmem.base) == -1)
	perror("shmdt");
    if (shmctl(Ownmem.shmid, IPC_RMID, NULL) == -1)
	perror("shmctl");
    if (semctl(Ownmem.semid, 0, IPC_RMID, NULL) == -1)
	perror("semctl");
#endif
    if (shmdt(Usock) == -1)
	perror("shmdt");
    exit(0);
}

/* Get user socket array */
sockinit()
{
	if(Usock != NULLUSOCK)
		return;	/* Already initialized */
	if ((Usockid = shmget(KEY_USOCK,0,0)) == -1)
	    perror("shmget KEY_USOCK");
	Usock = (struct usock *)shmat(Usockid, 0, 0);
}

/* Restore the "interrupts" */
void restore (mask)
char mask;
{
    struct sembuf sembuf;
    if (mask) {
	sembuf.sem_op = INC_SEM;
	sembuf.sem_num = SEM_MEMORY;
	sembuf.sem_flg = 0;
	semop(Mem->semid, &sembuf, 1);
	Intoff = 0;
    }
}
coreleft()
{
    return 0;
}

