#include <stdio.h>
#include <termio.h>
#include <signal.h>
#include "global.h"
#include <ctype.h>
#include "mbuf.h"
#include "socket.h"
#include "tty.h"
#define CMD_MODE        1       /* Command mode */
#define CONV_MODE       2       /* Converse mode */
static int Mode,Savedmode,Whoiswho;
static char Escape = 0x1d;	/* default escape character is ^] */
static char *Prog;
static FILE **Record;
static struct termio Savecon;
void reset_io();

main(argc,argv)
int argc;
char *argv[];
{
    int s;
    char *cp;
    Prog = argv[0];
    if (argc < 3) {
	fprintf(stderr,"Usage: %s <label> <callsign>\n",Prog);
	exit(1);
    }
    if (!strcmp(argv[1],"netrom")) {
	if ((s = nrcallsetup(argv[2])) == -1){
	    iostop();
	    exit(1);
	}
    }
    else
	if ((s = axcallsetup(argv[1],argv[2])) == -1){
	    iostop();
	    exit(1);
	}
    set_io();
    Record = (FILE **)shcalloc(1,sizeof(*Record));
    if ((Whoiswho = fork()) == 0) {
	sockreader(s);
	exit(0);
    }
    keyboard(s);
    cp = sockerr(s);
    printf("Connection closed: %s\n",cp != NULLCHAR ? cp : "EOF");
    close_s(s);
    reset_io();
}

/* Establish an AX.25 connection */
int
axcallsetup(label,call)
char *label, *call;
{
    struct sockaddr_ax fsocket;
    int s;

    if((s = socket(AF_AX25,SOCK_STREAM,0)) == -1){
	fprintf(stderr,"Can't create socket: errno %d\n",errno);
	return -1;
    }
    fsocket.sax_family = AF_AX25;
    setcall(&fsocket.ax25_addr,call);
    strncpy(fsocket.iface,label,ILEN);
    printf("Trying %s...\n",psocket((struct sockaddr *)&fsocket));
    if(connect(s,&fsocket,sizeof(fsocket)) == -1){
	printf("Connect failed: %s errno %d\n",sockerr(s),errno);
	close_s(s);
	return -1;
    }
    printf("Connected.\nEscape character is '^%c'.\n",Escape+'@');
    return s;
}

/* Initiate a NET/ROM transport connection */
int
nrcallsetup(call)
char *call;
{
	struct sockaddr_nr fsocket;
	int s;
	char alias[10];

	if((s = socket(AF_NETROM,SOCK_SEQPACKET,0)) == -1){
	    fprintf(stderr,"Can't create socket: errno %d\n",errno);
	    return -1;
	}
	/* Only the user callsign of the remote station is never used by */
	/* NET/ROM, but it is needed in the case of a psocket() call. */
	setcall((struct ax25_addr *)fsocket.nr_addr.user_call,call);
	setcall((struct ax25_addr *)fsocket.nr_addr.node_call,call);
	fsocket.nr_family = AF_NETROM;
	pax25(alias,(struct ax25_addr *)fsocket.nr_addr.node_call);
	printf("Trying %s...\n",alias);
	if(connect(s,&fsocket,sizeof(fsocket)) == -1){
	    printf("Connect failed: %s errno %d\n",sockerr(s),errno);
	    close_s(s);
	    return -1;
	}
	printf("Connected.\nEscape character is '^%c'.\n",Escape+'@');
	return s;
}

keyboard(s)
int s;
{
    struct mbuf *bp;
    int c;
    char *cp;

    Mode = CONV_MODE;
    ttysetmode(TTY_ECHO|TTY_EDIT);
    for(;;) {
	if ((c = getchar()) == EOF)
	    return;
	if (c == Escape && Mode == CONV_MODE) {
	    Savedmode = ttygetmode();
	    ttysetmode(TTY_ECHO|TTY_EDIT);
	    Mode = CMD_MODE;
	    printf("%s> ",Prog);
	    continue;
	}
	if ((bp = ttydriv(c)) != NULLBUF)
	    switch(Mode){
	    case CMD_MODE:
		if (keybcommand(bp,s) != 0)
		    return;
		if (Mode == CMD_MODE)
		    printf("%s> ",Prog);
		else
		    ttysetmode(Savedmode);
		break;
	    case CONV_MODE:
		/* If we're recording, record it */
		if(*Record != NULLFILE)
		    write_p(*Record,bp);

		/* Get rid of the trailing newline */
		if((cp = strchr(bp->data,'\n')) != NULLCHAR){
		    *cp = '\0';
		    bp->cnt--;
		}
		if (send_mbuf(s,bp,0,NULLCHAR,0) == -1)
		    return;
		break;
	    }
    }
}

sockreader(s)
int s;
{
    struct mbuf *bp;
    char c;

    for(;;){
	if(recv_mbuf(s,&bp,0,0,NULLCHAR,0) == -1)
	    break;
	while(pullup(&bp,&c,1) == 1){
	    /* Suspend output if we're not current */
/*	    while(Mode != CONV_MODE)
		;	*/		/* Busy wait ! */
	    putchar(c);
	    if(c == '\r')
		putchar('\n');

	    if(*Record != NULLFILE){
#ifdef UNIX
		if(c != '\r')
#endif				
		    putc(c,*Record);
		if(c == '\r')
		    putc('\n',*Record);
	    }
	}
    }
    /* Close seen from remote host */
}

/* parse commands, quick and dirty */
int
keybcommand(bp,s)
struct mbuf *bp;
int s;
{
    char buf[80], *p;

    dqdata(bp,buf,80);
    p = buf;
    while (*p == ' ' || *p == '\t')
	++p;
    if (isupper (*p))
	*p = tolower(*p);
    if (*p == '\n' || *p == '\r') {
	Mode = CONV_MODE;
	return 0;
    }
    if (*p == '?' || *p == 'h') {
	printf("Commands may be abbreviated.  Commands are:\n\n");
	printf("kick\t\tkick connection\n");
	printf("quit\t\texit program\n");
	printf("record <file>\trecord into <file>\n");
	printf("z\t\tsuspend program\n");
	return 0;
    }
    if (*p == 'k') {
	sockkick(s);
	Mode = CONV_MODE;
	return 0;
    }
    if (*p == 'z') {
#ifdef SYS5
	printf("Not yet implemented\n");
#else
	kill(getpid(),SIGSTOP);
	set_io();
	Mode = CONV_MODE;
#endif
	return 0;
    }
    if (*p == 'q') {
	return -1;
    }
    if (*p == 'r') {
	p = strchr(p,' ');
	if (p) {
	    if (!strcmp(p,"off")){
		fclose(*Record);
		*Record = NULLFILE;
	    }
	    else
		if ((*Record = fopen(p,"a")) == NULLFILE)
		    printf("Could not open %s for writing\n",p);
	}
	Mode = CONV_MODE;
	return 0;
    }
    printf("?Invalid command.\n");
    return 0;
}

set_io()
{
    struct termio ttybuf;

    /* We want unbuffered output */
    freopen(ttyname(1),"w",stdout);
    setbuf(stdout,NULLCHAR);
    ioctl(fileno(stdin), TCGETA, &ttybuf);
    Savecon = ttybuf;
    ttybuf.c_iflag &= ~IXON;
    ttybuf.c_lflag &= ~(ICANON|ECHO);
    ttybuf.c_cc[VTIME] = 0;
    ttybuf.c_cc[VMIN] = 1;
    ioctl(fileno(stdin), TCSETAW, &ttybuf);
    signal(SIGINT, reset_io);
    signal(SIGTERM, reset_io);
    signal(SIGHUP, reset_io);
    signal(SIGQUIT, reset_io);
}
void
reset_io()
{
    if (*Record != NULLFILE)
	fclose(*Record);
    if (Whoiswho != 0)
	shfree((char *)Record);
    ioctl(fileno(stdin), TCSETAW, &Savecon);
    iostop();
}

