/* FTP client (interactive user) code */
#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include "global.h"
#include "mbuf.h"
#include "cmdparse.h"
#include "socket.h"
#include "ftp.h"
#include "ftpcli.h"

#if	defined(__STDC__) || defined(__TURBOC__)
static int dotype(int argc,char *argv[]);
static int getsub(struct ftpcli *ftp,char *command,char *remotename,
	char *localname);
static void sendport(int s,struct sockaddr_in *socket);
#else
static int dotype(),getsub();
static void sendport();
#endif

static int donothing(),doftpcd(),dolist(),doget(),dols(),doput(),
	domkdir(),dormdir(),doascii(),dobinary(),doquit(),dohelp();

static char *Prog;
static char Notsess[] = "Not an FTP session!\n";
static struct ftpcli *ftp, *ftp_create();
static void reset_io();

struct cmds Ftpcmds[] = {
	"",		donothing,	0, 0, NULLCHAR,
	"ascii",	doascii,	0, 0, NULLCHAR,
	"binary",	dobinary,	0, 0, NULLCHAR,
	"cd",		doftpcd,	0, 2, "cd <directory>",
	"dir",		dolist,		0, 0, NULLCHAR,
	"list",		dolist,		0, 0, NULLCHAR,
	"get",		doget,		0, 2, "get remotefile <localfile>",
	"ls",		dols,		0, 0, NULLCHAR,
	"mkdir",	domkdir,	0, 2, "mkdir <directory>",
	"nlst",		dols,		0, 0, NULLCHAR,
	"quit",		doquit,		0, 0, NULLCHAR,
	"rmdir",	dormdir,	0, 2, "rmdir <directory>",
	"put",		doput,		0, 2, "put localfile <remotefile>",
	"type",		dotype,		0, 0, NULLCHAR,
	"?",		dohelp,		0, 0, NULLCHAR,
	NULLCHAR,	NULLFP,		0, 0, NULLCHAR,
};

main(argc,argv)
int argc;
char *argv[];
{
    int s;
    char *cp;
    Prog = argv[0];
    if (argc < 2) {
	fprintf(stderr,"Usage: %s <address> [<port>]\n",Prog);
	exit(1);
    }
    if ((s = callsetup(argc,argv)) == -1) {
	iostop();
	exit(1);
    }
    doftp(s);
    cp = sockerr(s);
    printf("Connection closed: %s\n",cp != NULLCHAR ? cp : "EOF");
    close_s(s);
    reset_io();
}

int
callsetup(argc,argv)
int argc;
char *argv[];
{
	struct sockaddr_in fsocket;
	char *cp;
	int control;

	fsocket.sin_family = AF_INET;
	if(argc < 3)
		fsocket.sin_port = IPPORT_FTP;
	else
		fsocket.sin_port = atoi(argv[2]);

	if((fsocket.sin_addr.s_addr = resolve(argv[1])) == 0){
		fprintf(stderr,"Bad hostname, %s\n",argv[1]);
		return -1;
	}
	/* Open the control connection */
	if((control = socket(AF_INET,SOCK_STREAM,0)) == -1){
		fprintf(stderr,"Can't create socket\n");
		return -1;
	}
	printf("Trying %s...\n",psocket((struct sockaddr *)&fsocket));
	if(connect(control,(char *)&fsocket,sizeof(fsocket)) == -1){
		cp = sockerr(control);
		printf("FTP connection failed: %s\n",
		 cp != NULLCHAR ? cp : "");
		return -1;
	}
	printf("Connected to %s.\n",argv[1]);

	/* Initialize shared FTP descriptor */
	ftp = ftp_create();
	ftp->control = control;

	return ftp->control;
}
doftp(control)
int control;
{
	int resp;
	char *cp;
	char *getpass();
	struct mbuf *bp,*bpsav;

	/* Wait for greeting from server */
	resp = getresp(control,200);

	/* Now process responses and commands */
	for(;;){
		switch(resp){
		case -1:
			return;
		case 220: /* Sign-on banner; prompt for and send USER command */
			bp = qdata("USER ",5);
			append(&bp,getline("Enter user name: "));
			if(send_mbuf(control,bp,0,NULLCHAR,0) == -1)
				return;
			resp = getresp(control,200);
			break;
		case 331: /* Password prompt; get password */
			bp = qdata("PASS ",5);
			/* Get the password without echoing */
			if((cp = getpass("Password: ")) == NULLCHAR)
			    return;
			append(&bp,qdata(cp,strlen(cp)));
			append(&bp,qdata("\r\n",2)); /* add CRNL */
			if(send_mbuf(control,bp,0,NULLCHAR,0) == -1)
				return;
			resp = getresp(control,200);
			break;
		default:
			/* Test the control channel first */
			if(sockstate(control) == NULLCHAR)
				return;

			bp = getline("ftp> ");

			/* Copy because cmdparse modifies the original */
			bpsav = copy_p(bp,len_mbuf(bp));
			if((resp = cmdparse(Ftpcmds,bp->data)) != -1){
				/* Valid command, free buffer and get another */
				free_p(bpsav);
			} else {
				/* Not a local cmd, send to remote server */
				if(send_mbuf(control,bpsav,0,NULLCHAR,0) == -1){
					free_p(bp);
					return;
				}
				resp = getresp(control,200);
			}
			free_p(bp);
			break;
		}
	}
}

/* Handle null line to avoid trapping on first command in table */
static
int
donothing(argc,argv)
int argc;
char *argv[];
{
	return 0;
}
/* Close session */
static int
doquit(argc,argv)
int argc;
char *argv[];
{
	usprintf(ftp->control,"QUIT\r\n");
	getresp(ftp->control,200);	/* Get the closing message */
	getresp(ftp->control,200);	/* Wait for the server to close */
	return -1;
}

static int
dohelp(argc,argv)
int argc;
char *argv[];
{
    int cnt = 0;
    char *p;
    printf("Local commands may be abbreviated.  Local commands are:\n\n");
    p = Ftpcmds[cnt].name;
    while (p) {
	++cnt;
	if ((cnt % 5) == 0)
	    printf("%-15s\n",p);
	else printf("%-16s",p);
	p = Ftpcmds[cnt].name;
    }
    return 0;
}

/* Translate 'cd' to 'cwd' for convenience */
static int
doftpcd(argc,argv)
int argc;
char *argv[];
{
	usprintf(ftp->control,"CWD %s\r\n",argv[1]);
	return getresp(ftp->control,200);
}
/* Translate 'mkdir' to 'xmkd' for convenience */
static int
domkdir(argc,argv)
int argc;
char *argv[];
{
	usprintf(ftp->control,"XMKD %s\r\n",argv[1]);
	return getresp(ftp->control,200);
}
/* Translate 'rmdir' to 'xrmd' for convenience */
static int
dormdir(argc,argv)
int argc;
char *argv[];
{
	usprintf(ftp->control,"XRMD %s\r\n",argv[1]);
	return getresp(ftp->control,200);
}
static int
dobinary(argc,argv)
int argc;
char *argv[];
{
	char *args[2];

	args[1] = "I";
	return dotype(2,args);
}
static int
doascii(argc,argv)
int argc;
char *argv[];
{
	char *args[2];

	args[1] = "A";
	return dotype(2,args);
}

/* Handle "type" command from user */
static int
dotype(argc,argv)
int argc;
char *argv[];
{
	if(argc < 2){
		switch(ftp->type){
		case IMAGE_TYPE:
			printf("Image\n");
			break;
		case ASCII_TYPE:
			printf("Ascii\n");
			break;
		case LOGICAL_TYPE:
			printf("Logical bytesize %u\n",ftp->logbsize);
			break;
		}
		return 0;
	}
	switch(*argv[1]){
	case 'i':
	case 'I':
	case 'b':
	case 'B':
		ftp->typesent = ftp->type = IMAGE_TYPE;
		usprintf(ftp->control,"TYPE I\r\n");
		break;
	case 'a':
	case 'A':
		ftp->typesent = ftp->type = ASCII_TYPE;
		usprintf(ftp->control,"TYPE A\r\n");
		break;
	case 'L':
	case 'l':
		ftp->typesent = ftp->type = LOGICAL_TYPE;
		ftp->logbsize = atoi(argv[2]);
		usprintf(ftp->control,"TYPE L %s\r\n",argv[2]);
		break;
	default:
		printf("Invalid type %s\n",argv[1]);
		return 1;
	}
	return getresp(ftp->control,200);
}
/* Start receive transfer. Syntax: get <remote name> [<local name>] */
static int
doget(argc,argv)
int argc;
char *argv[];
{
	char *remotename,*localname;
	if(ftp == NULLFTP){
		printf(Notsess);
		return 1;
	}
	remotename = argv[1];
	if(argc < 3)
		localname = remotename;
	else
		localname = argv[2];

	return getsub(ftp,"RETR",remotename,localname);
}
/* List remote directory. Syntax: dir <remote files> [<local name>] */
static int
dolist(argc,argv)
int argc;
char *argv[];
{
	char *remotename,*localname;
	if(ftp == NULLFTP){
		printf(Notsess);
		return 1;
	}
	remotename = argv[1];
	if(argc > 2)
		localname = argv[2];
	else
		localname = NULLCHAR;
	return getsub(ftp,"LIST",remotename,localname);
}
/* Remote directory list, short form. Syntax: ls <remote files> [<local name>]
 */
static int
dols(argc,argv,sp)
int argc;
char *argv[];
struct session *sp;
{
	char *remotename,*localname;
	if(ftp == NULLFTP){
		printf(Notsess);
		return 1;
	}
	remotename = argv[1];
	if(argc > 2)
		localname = argv[2];
	else
		localname = NULLCHAR;
	return getsub(ftp,"NLST",remotename,localname);
}
/* Common code to LIST/NLST/RETR */
static int
getsub(ftp,command,remotename,localname)
struct ftpcli *ftp;
char *command,*remotename,*localname;
{
	unsigned long total;
	FILE *fp;
	int cnt,resp,i,control;
	char *mode;
	struct sockaddr_in lsocket;
	time_t startclk;
	long rate;

	if(ftp == NULLFTP)
		return -1;
	control = ftp->control;

	switch(ftp->type){
	case IMAGE_TYPE:
	case LOGICAL_TYPE:
		mode = WRITE_BINARY;
		break;
	case ASCII_TYPE:
		mode = WRITE_TEXT;
		break;
	}
	/* Send TYPE message, if necessary */
	if(strcmp(command,"LIST") == 0 || strcmp(command,"NLST") == 0){
		if(ftp->typesent != ASCII_TYPE){
			/* Directory listings are always in ASCII */
			usprintf(control,"TYPE A\r\n");
			ftp->typesent = ASCII_TYPE;
			resp = getresp(control,200);
			if(resp == -1 || resp > 299){
				return 1;
			}
		}
	} else if(ftp->typesent != ftp->type){
		switch(ftp->type){
		case ASCII_TYPE:
			usprintf(control,"TYPE A\r\n");
			break;
		case IMAGE_TYPE:
			usprintf(control,"TYPE I\r\n");
			break;
		case LOGICAL_TYPE:
			usprintf(control,"TYPE L %d\r\n",ftp->logbsize);
			break;
		}
		ftp->typesent = ftp->type;
		resp = getresp(control,200);
		if(resp == -1 || resp > 299){
			return 1;
		}
	}
	if(localname == NULLCHAR){
		fp = stdout;
	} else if((fp = fopen(localname,mode)) == NULLFILE){
		printf("Can't write %s: %s\n",localname,sys_errlist[errno]);
		return 1;
	}
	/* Open the data connection */
	ftp->data = socket(AF_INET,SOCK_STREAM,0);
	listen(ftp->data,0);	/* Accept only one connection */
	ftp->state = RECEIVING_STATE;

	/* Send the PORT message and wait for ack */
	i = SOCKSIZE;
	getsockname(ftp->data,(char *)&lsocket,&i);
	sendport(control,&lsocket);
	resp = getresp(control,200);
	if(resp == -1 || resp > 299){
		/* Error, quit */
		if(fp != stdout)
			fclose(fp);
		close_s(ftp->data);
		ftp->data = -1;
		ftp->state = COMMAND_STATE;
		return 1;
	}
	/* Generate the command to start the transfer and wait for ack */
	if(remotename != NULLCHAR)
		usprintf(control,"%s %s\r\n",command,remotename);
	else
		usprintf(control,"%s\r\n",command);
	/* Get the intermediate "150" response */
	resp = getresp(control,100);
	if(resp == -1 || resp >= 400){
		/* Error, quit */
		if(fp != stdout)
			fclose(fp);
		close_s(ftp->data);
		ftp->data = -1;
		ftp->state = COMMAND_STATE;
		return 1;
	}
	/* Wait for the server to open the data connection */
	cnt = 0;
	ftp->data = accept(ftp->data,NULLCHAR,&cnt);
	startclk = time((time_t *)0);

	total = recvfile(fp,ftp->data,ftp->type);
	/* Immediately close the data connection; some servers (e.g., TOPS-10)
	 * wait for the data connection to close completely before returning
	 * the completion message on the control channel
	 */
	close_s(ftp->data);
	ftp->data = -1;

#ifdef	CPM
	if(ftp->type == ASCII_TYPE)
		fputc(CTLZ,fp);
#endif
	if(fp != stdout)
		fclose(fp);
	startclk = time((time_t *)0) - startclk;
	if(startclk != 0)
		rate = total/startclk;
	else
		rate = 0;
	if(total != -1)
		printf("Get complete: %lu bytes in %lu sec (%lu/sec)\n",
		 total,startclk,rate);
	else
		printf("Error or abort during data transfer\n");
	getresp(control,200);

	ftp->state = COMMAND_STATE;
	return 0;
}
/* Send a file. Syntax: put <local name> [<remote name>] */
static int
doput(argc,argv)
int argc;
char *argv[];
{
	char *remotename,*localname,*mode;
	int i,resp,control;
	unsigned long total;
	FILE *fp;
	struct sockaddr_in lsocket;
	time_t startclk;
	long rate;

	if(ftp == NULLFTP){
		printf(Notsess);
		return 1;
	}
	control = ftp->control;

	localname = argv[1];
	if(argc < 3)
		remotename = localname;
	else
		remotename = argv[2];

	if(ftp->type == IMAGE_TYPE)
		mode = READ_BINARY;
	else
		mode = READ_TEXT;

	/* Send TYPE message, if necessary */
	if(ftp->typesent != ftp->type){
		switch(ftp->type){
		case ASCII_TYPE:
			usprintf(control,"TYPE A\r\n");
			break;
		case IMAGE_TYPE:
			usprintf(control,"TYPE I\r\n");
			break;
		case LOGICAL_TYPE:
			usprintf(control,"TYPE L %d\r\n",ftp->logbsize);
			break;
		}
		ftp->typesent = ftp->type;
		resp = getresp(control,200);
		if(resp == -1 || resp > 299){
			return 1;
		}
	}
	if((fp = fopen(localname,mode)) == NULLFILE){
		printf("Can't read %s: %s\n",localname,sys_errlist[errno]);
		return 1;
	}
	/* Open the data connection */
	ftp->data = socket(AF_INET,SOCK_STREAM,0);
	listen(ftp->data,0);

	ftp->state = SENDING_STATE;

	/* Send the PORT message and wait for ack */
	i = SOCKSIZE;
	getsockname(ftp->data,(char *)&lsocket,&i);
	sendport(control,&lsocket);
	resp = getresp(control,200);
	if(resp == -1 || resp > 299){
		/* Error, quit */
		fclose(fp);
		close_s(ftp->data);
		ftp->data = -1;
		ftp->state = COMMAND_STATE;
		return 1;
	}
	/* Generate the command to start the transfer and wait for ack */
	usprintf(control,"STOR %s\r\n",remotename);
	resp = getresp(control,100);
	if(resp == -1 || resp >= 400){
		/* Error, quit */
		fclose(fp);
		close_s(ftp->data);
		ftp->data = -1;
		ftp->state = COMMAND_STATE;
		return 1;
	}
	/* Wait for the data connection to open. Otherwise the first
	 * block of data would go out with the SYN, and this may confuse
	 * some other TCPs
	 */
	accept(ftp->data,NULLCHAR,(int *)NULL);

	startclk = time((time_t *)0);

	total = sendfile(fp,ftp->data,ftp->type);
	close_s(ftp->data);
	ftp->data = -1;
	fclose(fp);

	startclk = time((time_t *)0) - startclk;
	if(startclk != 0)
		rate = total/startclk;
	else
		rate = 0;
	if(total != -1)
		printf("Put complete: %lu bytes in %lu sec (%lu/sec)\n",
		 total,startclk,rate);
	else
		printf("Error or abort during data transfer\n");

	getresp(control,200);
	ftp->state = COMMAND_STATE;
	return 1;
}
/* Abort a GET or PUT operation in progress. Note: this will leave
 * the partial file on the local or remote system
 */
static int
doabort(argc,argv)
int argc;
char *argv[];
{
	switch(ftp->state){
	case COMMAND_STATE:
		printf("No active transfer\n");
		return 1;
	case SENDING_STATE:
		/* Send a premature EOF.
		 * Unfortunately we can't just reset the connection
		 * since the remote side might end up waiting forever
		 * for us to send something.
		 */
		shutdown(ftp->data,1);
		break;
	case RECEIVING_STATE:
		/* Just blow away the receive socket */
		shutdown(ftp->data,2);
		break;
	}
	return 0;
}
/* send PORT message */
static void
sendport(s,socket)
int s;
struct sockaddr_in *socket;
{
	struct mbuf *bp;

	/* Compose and send PORT a,a,a,a,p,p message */
	if((bp = alloc_mbuf(35)) == NULLBUF){	/* 5 more than worst case */
		printf("No space!\n");
		return;
	}
	/* I know, this looks gross, but it works! */
	sprintf(bp->data,"PORT %u,%u,%u,%u,%u,%u\r\n",
		hibyte(hiword(socket->sin_addr.s_addr)),
		lobyte(hiword(socket->sin_addr.s_addr)),
		hibyte(loword(socket->sin_addr.s_addr)),
		lobyte(loword(socket->sin_addr.s_addr)),
		hibyte(socket->sin_port),
		lobyte(socket->sin_port));
	bp->cnt = strlen(bp->data);
	send_mbuf(s,bp,0,NULLCHAR,0);
}

/* Wait for, read and display response from FTP server. Return the result code.
 */
int
getresp(s,mincode)
int s;
int mincode;	/* Keep reading until at least this code comes back */
{
	register char *line;
	int rval;

	line = malloc(256);
	for(;;){
		/* Get line */
		if(recvbuf(s,line,256) == -1){
			rval = -1;
			break;
		}
		rip(line);		/* Remove cr/lf */
		printf("%s\n",line);	/* Display to user */

		/* Messages with dashes are continued */
		if(line[3] != '-' && (rval = atoi(line)) >= mincode)
			break;
	}
	free(line);
	return rval;
}

/* Issue a prompt and read a line from the user */
struct mbuf *
getline(prompt)
char *prompt;
{
    char buf[258];
    printf("%s",prompt);
    if (gets(buf) == NULLCHAR)
	return NULLBUF;
    strcat(buf,"\r\n");
    return qdata(buf,strlen(buf));
}
/* Allocate an FTP client control block */
static struct ftpcli *
ftp_create()
{
	register struct ftpcli *p;

	if((p = (struct ftpcli *)shcalloc(1,sizeof (struct ftpcli))) == NULLFTP)
		return NULLFTP;

	p->state = COMMAND_STATE;
        p->type = ASCII_TYPE;		/* Default transfer type */
	p->typesent = ASCII_TYPE;	/* Server default is ascii */
	p->control = p->data = -1;
	return p;
}

/* This is like recvline() but it tries to grab as much as it needs at once,
   more like recvchar(), minimizing the number of rpc calls. NB: The data is
   stored in a static mbuf. EOL support for AX25 and NETROM is easily added
   if necessary.
 */
int
recvbuf(s,buf,len)
int s;		/* Socket index */
char *buf;	/* User buffer */
unsigned len;	/* Length of buffer */
{
	char c;
	int cnt = 0;
	static struct mbuf *bp = NULLBUF;

	while(len-- != 0){
		if(pullup(&bp,&c,1) != 1){
			if (recv_mbuf(s,&bp,len+1,0,NULLCHAR,0) < 1){
				*buf = '\0';
				return -1;
			}
			pullup(&bp,&c,1);
		}
		*buf++ = c;
		cnt++;
		if(uchar(c) == '\n')
			break;
	}
	if(len != 0)
		*buf = '\0';
	return cnt;
}
set_io()
{
    signal(SIGINT, reset_io);
    signal(SIGTERM, reset_io);
    signal(SIGHUP, reset_io);
    signal(SIGQUIT, reset_io);
}
static void
reset_io()
{
    if(ftp->fp != NULLFILE && ftp->fp != stdout)
	fclose(ftp->fp);
    close_s(ftp->data);
    shfree((char *)ftp);
    iostop();
}
