#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include "global.h"
#include "config.h"
#include "sems.h"
#include "mbuf.h"
#include "memory.h"
#include "timer.h"
#include "usock.h"
#include "slip.h"
#include "unix.h"

char *shmalloc();
struct mbuf **Hopper;
struct memory Mainmem;
int *Selectcnt;
extern struct memory *Mem;
int Usockid;
int Intoff=0, Inputpending=0, Ownpid, Tick;
init()
{
    int i;
    void iostop(), timerint();
    struct sembuf sembuf;
    struct mrec *mrec;

    signal(SIGINT, iostop);
    signal(SIGTERM, iostop);
    signal(SIGHUP, iostop);
    signal(SIGQUIT, iostop);
    signal(SIGALRM, timerint);

    if ((Mainmem.semid = semget(KEY_CNTL, SEMNOS, 0666 | IPC_CREAT)) == -1) {
	perror("semget");
#ifdef SEMMSL
	if (SEMMSL < SEMNOS) {
	    printf("%s semaphores are needed, but your computer has a maximum\n",SEMNOS);
	    printf("of %s semaphores. You will have to change either DEFNSOCK\n",SEMMSL);
	    printf("in config.h or SEMMSL in your kernel.\n");
	}
#else
	printf("The problem might be that your system cannot give me %s semaphores.\n",
	       SEMNOS);
	printf("Try changing DEFNSOCK in config.h or alter the configuration of your kernel.\n");
#endif
    }

    for (i=0; i<SEMNOS; ++i) {  /* initialize all semaphores to 1 */
	sembuf.sem_num = i;
	sembuf.sem_op = TEST0_SEM;
	sembuf.sem_flg = IPC_NOWAIT;
	if (semop(Mainmem.semid, &sembuf, 1) == 0) {
	  sembuf.sem_op = INC_SEM;
	  semop(Mainmem.semid, &sembuf, 1);
	}
    }
    sembuf.sem_num = SEM_SELECT; /* this one, however, should start as zero */
    sembuf.sem_op = WAITDEC_SEM;
    sembuf.sem_flg = 0;
    semop(Mainmem.semid, &sembuf, 1);

    Mainmem.block = 0;
    Mainmem.next = NULLMEM;
    grabcore(&Mainmem,200000,0,KEY_BUFFER);
    Mainmem.queue = (struct mbuf **) shcalloc(1,sizeof(*Mainmem.queue));
    Hopper = (struct mbuf **) shcalloc(1,sizeof(*Hopper));
    Selectcnt = (int *) shmalloc(sizeof(int));
    *Selectcnt = 0;
    
    sockinit();
    
    /* Store the addresses of our memory segments */
    mrec = (struct mrec *)Mainmem.shmbase;
    mrec->data[0] = (int32) Mainmem.shmbase;
    mrec->data[1] = (int32) Usock;
    mrec->data[2] = (int32) Mainmem.queue;
    mrec->data[3] = (int32) Hopper;
    mrec->data[4] = (int32) Selectcnt;

    Ownpid = getpid();
    /* We will get a SIGALRM signal at every MSPTICK milliseconds.
       You may try to tweak its definiton in timer.h */
    ualarm (100000,MSPTICK*1000);
}

void
iostop()
{
    struct memory *mp;
    int i;
    for(i=0; i<Nasy; ++i)
	if (Asy[i].pid != -1)
	    kill(Asy[i].pid,SIGTERM);
    if (shmdt(Usock) == -1)
	perror("shmdt");
    if (shmctl(Usockid, IPC_RMID, NULL) == -1)
	perror("shmctl");
    mp = &Mainmem;
    while (mp != NULLMEM) {
	if (shmdt(mp->shmbase) == -1)
	    perror("shmdt");
	if (shmctl(mp->shmid, IPC_RMID, NULL) == -1)
	    perror("shmctl");
	if (semctl(mp->semid, 0, IPC_RMID, NULL) == -1)
	    perror("semctl");
	mp = mp->next;
    }
    exit(0);
}

/* Initialize user socket array */
sockinit()
{
	if(Usock != NULLUSOCK)
		return;	/* Already initialized */
	if ((Usockid = shmget(KEY_USOCK,Nusock*sizeof(struct usock),
			      0666 | IPC_CREAT)) == -1)
	    perror("shmget KEY_USOCK");
	Usock = (struct usock *)shmat(Usockid, 0, 0);
}

void timerint()
{
    signal(SIGALRM, timerint);
    ++Tick;
}

/* Restore the "interrupts" */
void restore (mask)
    char mask;
{
    struct sembuf sembuf;
    if (mask) {
	sembuf.sem_op = INC_SEM;
	sembuf.sem_num = SEM_MEMORY;
	sembuf.sem_flg = 0;
	semop(Mem->semid, &sembuf, 1);
	Intoff = 0;
	/* If we got a SIGPOLL when the interrupts where disabled, take care
	   of it now, by reissueing it. */
	if (Inputpending)
	    kill (Ownpid, SIGPOLL); /* Argh */
    }
}

systick()
{
}

coreleft()
{
    return 0;
}

int32
initclimem(key)
key_t key;
{
    struct memory *mp;
    struct mrec *mrec;
    struct sembuf sembuf[2];
    int i;
    mp = &Mainmem;
    while (mp->next != NULLMEM)
	mp = mp->next;
    mp->next = (struct memory *)malloc(sizeof(struct memory));
    mp = mp->next;
    mp->next = NULLMEM;
    mp->block = 0;
    if ((mp->semid = semget(key, 2, 0666 | IPC_CREAT)) == -1)
	perror("semget");

    for (i=0; i<2; ++i) {
	sembuf[i].sem_num = i;
	sembuf[i].sem_op = INC_SEM;
	sembuf[i].sem_flg = 0;
    }
    semop(mp->semid, sembuf, 2); /* initialize all semaphores to 1 */
    grabcore(mp,30000,0,key);
    mp->queue = (struct mbuf **)shmalloc(sizeof(*mp->queue));
    *mp->queue = NULLBUF;
    mrec = (struct mrec *)mp->shmbase;
    mrec->data[2] = (int32) mp->queue; /* store the address of the queue */
    Mem = &Mainmem; /* restore our default memory */
    return (int32)mp->shmbase;
}

asy_rxstart(dev)
int16 dev;
{
    char buf[80], *p1, *p2;

    sprintf(buf,"%d",Asy[dev].IORser);
    p1 = buf + strlen(buf) + 1;
    sprintf(p1,"%d",Slip[dev].iface);
    p2 = p1 + strlen(p1) + 1;
    sprintf(p2,"%d",Slip[dev].type);
    signal(SIGCLD,SIG_IGN);
    if ((Asy[dev].pid = fork()) == 0) {
	execlp("asy_rx","asy_rx",buf,p1,p2,NULLCHAR);
	printf("Cannot execute 'asy_rx', check your path!\n");
    }
    if (Asy[dev].pid == -1)
    	printf("Could not fork the receiver process.\n");
}

#ifdef SYS5
/* System V release 4 or later may have ualarm(), then we won't need
   to do this mess. */
ualarm(initial,usecs)
long initial, usecs;
{
    int child;
    static char buf[10];
    sprintf(buf,"%ld",usecs/1000);
    signal(SIGCLD,SIG_IGN);
    if ((child = fork()) == 0) {
	sleep(initial/1000000); /* usleep() should rather be used here */
	execlp("nostimer","nostimer",buf,NULLCHAR);
	printf("Cannot execute 'nostimer', check your path!\n");
	exit(1);
    }
    if (child == -1)
    	printf("Could not fork the timer process.\n");
}
int
killpg(pid,sig)
int pid, sig;
{
    return kill(pid,sig);
}
#endif
