#include <stdio.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include "global.h"
#include "mbuf.h"
#include "config.h"
#include "sems.h"
#include "memory.h"
#include "cmdparse.h"
#include "iface.h"
#include "usock.h"

#ifdef	TRACE
#include "trace.h"
/* Dummy structure for loopback tracing */
struct iface Loopback = { NULLIF, "loopback" };
#endif
char Badhost[] = "Unknown host %s\n";
char Nospace[] = "No space!!\n";	/* Generic malloc fail message */
char Notval[] = "Not a valid control block\n";

extern int Tick;
extern struct memory Mainmem;
extern struct mbuf **Hopper;
extern char *Startup;	/* File to read startup commands from */
int doarp(),doax25(),doattach(),doip(),domode(),donetrom(),doroute(),dotrace(),
doparam(),dotcp();
struct cmds Cmds[] = {
#if	(defined(AX25) || defined(ETHER) || defined(APPLETALK))
	"arp",		doarp,		0, 0, NULLCHAR,
#endif
#ifdef	AX25
	"ax25",		doax25,		0, 0, NULLCHAR,
#endif	
	"attach",	doattach,	0, 2,
		"attach <hardware> <hw specific options>",
	"ip",		doip,		0, 0, NULLCHAR,
#ifdef	AX25
	"mode",		domode,		0, 2, "mode <interface>",
#endif
#ifdef	NETROM
	"netrom",	donetrom,	0, 0, NULLCHAR,
#endif
	"param",	doparam,	0, 2, "param <interface>",
	"route",	doroute,	0, 0, NULLCHAR,
	"tcp",		dotcp,		0, 0, NULLCHAR,
#ifdef TRACE
	"trace",	dotrace,	0, 0, NULLCHAR,
#endif
	NULLCHAR,	NULLFP,		0, 0,
		"Unknown command; type \"?\" for list",
};

main(argc,argv)
int argc;
char *argv[];
{
    FILE *fp;
    char *inbuf, *intmp;
    int i;
    struct sembuf sembuf;
    init();
	if(argc > 1){
		/* Read startup file named on command line */
		if((fp = fopen(argv[1],READ_TEXT)) == NULLFILE)
			perror("fopen config file");
	} else {
		fp = fopen(Startup,READ_TEXT);
	}
	if(fp != NULLFILE){
		inbuf = malloc(BUFSIZ);
		intmp = malloc(BUFSIZ);
		while(fgets(inbuf,BUFSIZ,fp) != NULLCHAR){
			strcpy(intmp,inbuf);
			if(cmdparse(Cmds,inbuf,NULL) != 0){
				printf("input line: %s",intmp);
			}
		}
		fclose(fp);
		free(inbuf);
		free(intmp);
	}

    sembuf.sem_num = SEM_RPC;
    sembuf.sem_flg = 0;
    while(1) {
	if (Tick)
	    timerproc();
        if(len_q(*Hopper) > 0)
	    network();
	sembuf.sem_op = TEST0_SEM;
	/* Wait until we get a rpc indication */
	i = semop(Mainmem.semid, &sembuf, 1);
	checksocks();
	if(i != -1) {
	    /* done only if we were not interrupted */
	    grabinput();
	    if(len_q(*Hopper) > 0)
		network();
	    sembuf.sem_op = INC_SEM;
	    semop(Mainmem.semid, &sembuf, 1); /* Now others may write again */
	}
    }
}
/*
struct mbuf *
fifotombuf(fifo, len)
struct fifo *fifo;
unsigned short len;
{
    struct mbuf *bp;
    unsigned short n, n2;

    n = min(len, fifo->cnt);
    bp = alloc_mbuf((int16)n);
    fifo->cnt -= n;
    bp->cnt = n;
    n2 = min(n,&fifo->buf[fifo->buflen] - fifo->rp);
    memcpy(bp->data, fifo->rp, n2);
    fifo->rp += n2;
    if (fifo->rp >= &fifo->buf[fifo->buflen])
	fifo->rp = fifo->buf;
    if (n2 == n)
	return bp;
    memcpy(bp->data + n2, fifo->rp, n - n2);
    return bp;
}
*/
/* List of supported hardware devices */
int asy_attach(),nr_attach();

struct cmds Attab[] = {
#ifdef	ASY
	/* Ordinary PC asynchronous adaptor */
	"asy", asy_attach, 0, 8, 
	"attach asy <address> <vector> slip|ax25|nrs <label> <buffers> <mtu> <speed>",
#endif
#ifdef NETROM
	/* fake netrom interface */
	"netrom", nr_attach, 0, 1,
	"attach netrom",
#endif
	NULLCHAR, NULLFP, 0, 0,
	"Unknown device",
};

/* Attach an interface
 * Syntax: attach <hw type> <I/O address> <vector> <mode> <label> <bufsize> [<speed>]
 */
int
doattach(argc,argv,envp)
int argc;
char *argv[];
void *envp;
{
	return subcmd(Attab,argc,argv,envp);
}
/* Manipulate I/O device parameters */
int
doparam(argc,argv)
int argc;
char *argv[];
{
	register struct iface *ifp;

	for(ifp=Ifaces;ifp != NULLIF;ifp = ifp->next){
		if(strcmp(argv[1],ifp->name) == 0)
			break;
	}
	if(ifp == NULLIF){
		printf("Interface \"%s\" unknown\n",argv[1]);
		return 1;
	}
	if(ifp->ioctl == NULLFP){
		printf("Not supported\n");
		return 1;
	}
	/* Pass rest of args to device-specific code */
	return (*ifp->ioctl)(ifp,argc-2,argv+2);
}
/* Process packets in the Hopper */
network()
{
    struct mbuf *bp;
    struct phdr phdr;

    while(len_q(*Hopper) > 0) {
	/* Process the input packet */
	bp = dequeue(Hopper);
	pullup(&bp,(char *)&phdr,sizeof(phdr));
	dump(phdr.iface,IF_TRACE_IN,phdr.type,bp);
	switch(phdr.type){
#ifdef	AX25
	case TYPE_KISS:
		/* Toss the type field and handle as AX.25 */
		pullup(&bp,NULLCHAR,1);	/*note fall-thru */
	case TYPE_AX25:
		ax_recv(phdr.iface,bp);
		break;
#endif
	case TYPE_ETHER:
		eproc(phdr.iface,bp);
		break;
	case TYPE_IP:
		ip_route(bp,0);
		break;
	case TYPE_APPLETALK:
	default:
		free_p(bp);
		break;
	}
    }
}

/* Protocol tracing function pointers */
#ifdef	TRACE
int ax25_dump(),ether_dump(),ip_dump(),at_dump(),ki_dump();

int (*Tracef[])() = {
#ifdef	AX25
	ax25_dump,
#else
	NULLFP,
#endif

#ifdef	ETHER
	ether_dump,
#else
	NULLFP,
#endif
	ip_dump,

#ifdef	APPLETALK
	at_dump,
#else
	NULLFP,
#endif

#ifdef	KISS
	ki_dump,
#else
	NULLFP,
#endif
};
#else	/* TRACE */
int (*Tracef[])() = { NULLFP };	/* No tracing at all */
dump(iface,direction,type,bp)
struct iface *iface;
int direction;
unsigned type;
struct mbuf *bp;
{
}
#endif	/* TRACE */

#ifdef	TRACE
int
dotrace(argc,argv)
int argc;
char *argv[];
{
	struct iface *ifp;

	if(argc < 2){
		showtrace(&Loopback);
		for(ifp = Ifaces; ifp != NULLIF; ifp = ifp->next)
			showtrace(ifp);
		return 0;
	}
	if(strcmp("loopback",argv[1]) == 0)
		ifp = &Loopback;
	else 
		for(ifp = Ifaces; ifp != NULLIF; ifp = ifp->next)
			if(strcmp(ifp->name,argv[1]) == 0)
				break;

	if(ifp == NULLIF){
		printf("Interface %s unknown\n",argv[1]);
		return 1;
	}
	if(argc >= 3)
		ifp->trace = htoi(argv[2]);

	showtrace(ifp);
	return 0;
}
/* Display the trace flags for a particular interface */
showtrace(ifp)
register struct iface *ifp;
{
	if(ifp == NULLIF)
		return;
	printf("%s:",ifp->name);
	if(ifp->trace & (IF_TRACE_IN | IF_TRACE_OUT)){
		if(ifp->trace & IF_TRACE_IN)
			printf(" input");
		if(ifp->trace & IF_TRACE_OUT)
			printf(" output");

		if(ifp->trace & IF_TRACE_HEX)
			printf(" (Hex/ASCII dump)");
		else if(ifp->trace & IF_TRACE_ASCII)
			printf(" (ASCII dump)");
		else
			printf(" (headers only)");
		printf("\n");
	} else
		printf(" tracing off\n");
}
#endif	/* TRACE */

#ifndef	NETROM
#ifdef	AX25
struct ax25_addr Nr_nodebc;
#endif	/* AX25 */
nr_route(bp)
struct mbuf *bp;
{
	free_p(bp);
}
nr_nodercv(bp)
struct mbuf *bp;
{
	free_p(bp);
}
#endif	/* NETROM */

/* Display or set IP interface control flags */
domode(argc,argv)
int argc;
char *argv[];
{
	register struct iface *ifp;

	for(ifp=Ifaces;ifp != NULLIF;ifp = ifp->next){
		if(strcmp(argv[1],ifp->name) == 0)
			break;
	}
	if(ifp == NULLIF){
		printf("Interface \"%s\" unknown\n",argv[1]);
		return 1;
	}
	if(argc < 3){
		printf("%s: %s\n",ifp->name,
		 (ifp->flags & CONNECT_MODE) ? "VC mode" : "Datagram mode");
		return 0;
	}
	switch(argv[2][0]){
	case 'v':
	case 'c':
	case 'V':
	case 'C':
		ifp->flags = CONNECT_MODE;
		break;
	case 'd':
	case 'D':
		ifp->flags = DATAGRAM_MODE;
		break;
	default:
		printf("Usage: %s [vc | datagram]\n",argv[0]);
		return 1;
	}
	return 0;
}
/* Note: make sure that there is no previous definition of tolower() */
char tolower (c)
    char c;
{
    return ((c >= 'A' && c <= 'Z') ? c + 'a' - 'A' : c);
}
int stxrdy(dev)
int dev;
{
    return 1;
}

/* Close any orphan sockets */
checksocks()
{
    int i, j, found;
    struct usock *up;
    struct sembuf sembuf;
    sembuf.sem_num = SEM_USOCK;
    sembuf.sem_flg = IPC_NOWAIT;
    sembuf.sem_op = WAITDEC_SEM;
    if (semop(Mainmem.semid, &sembuf, 1) == -1)
	 return;	/* would block */
    
    for(i=0; i < Nusock; ++i) {
	up = &Usock[i];
	if(up->type == NOTUSED)
	    continue;
	for(found = j = 0; j < sizeof(up->pid)/sizeof(int); ++j)
	    if(up->pid[j] && (kill(up->pid[j],0) != -1 || errno != ESRCH))
		++found;
	    else
		up->pid[j] = 0;
	if(!found) {
	    releaseusock();
	    close_s(i+USOCKBASE);
	    if (semop(Mainmem.semid, &sembuf, 1) == -1)
		 return;	/* would block */
        }
    }
    releaseusock();
}
