#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include "global.h"
#include "icmp.h"
#include "mbuf.h"
#include "socket.h"

char *inet_ntoa();
struct mbuf *htonicmp();
void pingtx(),reset_io();
static int pingem();
static int Whoiswho = 1;
static struct ping *Ping;
/* Send ICMP Echo Request packets */
main(argc,argv)
int argc;
char *argv[];
{
    int s;
    if (argc < 2) {
	fprintf(stderr,"Usage: %s <address> [<length>] [<interval>]\n",
		argv[0]);
	exit(1);
    }
    if ((s = callsetup(argc,argv)) == -1) {
	iostop();
	exit(1);
    }
    set_io();
    if(Ping->interval != 0){
	if ((Whoiswho = fork()) == 0) {
	    pingtx(s);
	    exit(0);
	}
    }
    else {
	Ping->sent = 1;
	pingem(s,0,(int16)getpid(),Ping->len);
    }
    recvping(s);
    close_s(s);
    reset_io();
}

callsetup(argc,argv)
int argc;
char *argv[];
{
	struct sockaddr_in to;
	int s;

	if((s = socket(AF_INET,SOCK_RAW,IPPROTO_ICMP)) == -1){
		fprintf(stderr,"Can't create socket\n");
		return -1;
	}
	to.sin_family = AF_INET;
	if((to.sin_addr.s_addr = resolve(argv[1])) == 0){
		fprintf(stderr,"Host %s unknown\n",argv[1]);
		close_s(s);
		return -1;
	}
	connect(s,(char *)&to,sizeof(to));

	Ping = (struct ping *) shcalloc(1,sizeof(struct ping));
	if(argc > 2)
		Ping->len = atoi(argv[2]);

	if(argc > 3)
		Ping->interval = atoi(argv[3]);
/*	else
		Ping->interval = 1; */
	
	return s;
}
recvping(s)
int s;
{
	struct sockaddr_in to, from;
	int len;
	struct mbuf *bp;
	struct icmp icmp;
	struct timeb timestamp, timenow;
	long rtt, abserr;

	len = sizeof(struct sockaddr_in);
	getpeername(s,&to,&len);
	/* Collect the replies */
	for(;;){
		len = sizeof(struct sockaddr_in);
		if(recv_mbuf(s,&bp,0,0,(char *)&from,&len) == -1)
			break;
		ntohicmp(&icmp,&bp);
		if(icmp.type != ECHO_REPLY
		 || from.sin_addr.s_addr != to.sin_addr.s_addr
		 || icmp.args.echo.id != (int16)getpid()){
			/* Ignore other people's responses */
			free_p(bp);
			continue;
		}
		Ping->responses++;
		pullup(&bp,(char *)&timestamp,sizeof(struct timeb));
		free_p(bp);

		/* Compute round trip time, update smoothed estimates */
		ftime(&timenow);
		rtt = (timenow.time - timestamp.time)*1000 +
		    timenow.millitm - timestamp.millitm;
		if(Ping->interval == 0){
		    printf("%s: rtt %lu",inet_ntoa(from.sin_addr.s_addr),rtt);
		    break;
		}
		abserr = rtt > Ping->srtt ? rtt - Ping->srtt: Ping->srtt - rtt;
		if(Ping->responses == 1){
			/* First response, base entire SRTT on it */
			Ping->srtt = rtt;
		} else {
			Ping->srtt = (7*Ping->srtt + rtt + 4) >> 3;
		}
		Ping->mdev = (3*Ping->mdev + abserr + 2) >> 2;

	}
	printf("\n");
}
/* Ping transmit process. Runs until socket is closed */
void
pingtx(s)
int s;		/* Socket to use */
{
	struct sockaddr_in sock;
	int len = sizeof(struct sockaddr_in);
	getpeername(s,&sock,&len);
	Ping->sent = 0;
	for(;;){
	    printf("pinging %s: sent %lu rcvd %lu avg rtt %lu mdev %lu    \r",
		   inet_ntoa(sock.sin_addr.s_addr),Ping->sent,Ping->responses,
		   Ping->srtt,Ping->mdev);
	    if(pingem(s,(int16)Ping->sent++,(int16)getppid(),Ping->len) == -1)
		   break;
	    sleep(Ping->interval);
	}
}

/* Send ICMP Echo Request packet */
static
pingem(s,seq,id,len)
int s;		/* Raw socket on which to send ping */
int16 seq;	/* ICMP Echo Request sequence number */
int16 id;	/* ICMP Echo Request ID */
int16 len;	/* Length of optional data field */
{
	struct mbuf *data = NULLBUF;
	struct mbuf *bp;
	struct icmp icmp;
	struct timeb timestamp;

	if((data = alloc_mbuf(len+sizeof(struct timeb))) == NULLBUF)
		return -1;
	data->cnt = len+sizeof(struct timeb);
	/* Set optional data field, if any, to all 55's */
	if(len != 0)
		memset(data->data+sizeof(struct timeb),0x55,len);

	/* Insert timestamp and build ICMP header */
	ftime(&timestamp);
	memcpy(data->data,(char *)&timestamp,sizeof(struct timeb));
/*	Icmp_stats.output[ECHO]++;*/
	icmp.type = ECHO;
	icmp.code = 0;
	icmp.args.echo.seq = seq;
	icmp.args.echo.id = id;
	if((bp = htonicmp(&icmp,data)) == NULLBUF){
		free_p(data);
		return -1;
	}
	if(send_mbuf(s,bp,0,NULLCHAR,0) == -1){
		free_p(bp);
		return -1;
	}
}

set_io()
{
    signal(SIGINT, reset_io);
    signal(SIGTERM, reset_io);
    signal(SIGHUP, reset_io);
    signal(SIGQUIT, reset_io);
}

void
reset_io()
{
    if(Whoiswho != 0)
	shfree((char *)Ping);
    iostop();
}
