#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/param.h>
#include "global.h"
#include "mbuf.h"
#include "sems.h"
#include "usock.h"
#include "memory.h"
#include "timer.h"
    
extern struct memory Mainmem, *Mem;
extern int Intoff;

/* Disable all "interrupts", don't mask any signals however */
int dirps()
{
    struct sembuf sembuf;
    if (Intoff)
	return 0;
    sembuf.sem_op = WAITDEC_SEM;
    sembuf.sem_num = SEM_MEMORY;
    sembuf.sem_flg = 0;
    while (semop(Mem->semid, &sembuf, 1) == -1 && errno == EINTR)
	;
    Intoff = 1;
    return 1;
}

int
pwait(usaddr)
struct usock *usaddr;
{
    int num, ret;
    struct sembuf sembuf;
    if (usaddr == NULLUSOCK)
	return 0;
    num = (usaddr - Usock) + SEM_SOCKET;
    sembuf.sem_num = SEM_USOCK;
    sembuf.sem_flg = 0;
    sembuf.sem_op = WAITDEC_SEM;
    while (semop(Mainmem.semid, &sembuf, 1) == -1 && errno == EINTR)
	;
#ifdef notyet
    if (usaddr->cb.p != NULLCHAR)	/* please note! */
	return 0;
#endif
    usaddr->waitcnt++;
    sembuf.sem_num = num;
    sembuf.sem_flg = IPC_NOWAIT;
    sembuf.sem_op = WAITDEC_SEM; /* Make sure it is zero */
    semop(Mainmem.semid, &sembuf, 1);
    sembuf.sem_num = SEM_USOCK;
    sembuf.sem_op = INC_SEM;
    semop(Mainmem.semid, &sembuf, 1);
    sembuf.sem_num = num;
    sembuf.sem_op = WAITDEC_SEM;
    sembuf.sem_flg = 0; /* now wait */
    while (semop(Mainmem.semid, &sembuf, 1) == -1 && errno == EINTR)
	;
    ret = usaddr->retval;
    return ret;
}

void
psignal(usaddr,n)
struct usock *usaddr;
int n;    
{
    int num, i, err;
    struct sembuf sembuf;
    extern int *Selectcnt;
    if (usaddr == NULLUSOCK)
	return;
    num = (usaddr - Usock) + SEM_SOCKET;
    holdusock();
    n = min(n,usaddr->waitcnt);
    if (n == 0)
	n = usaddr->waitcnt;
    usaddr->retval = 0;
    sembuf.sem_num = num;
    while (n) {
	sembuf.sem_op = TEST0_SEM;
	sembuf.sem_flg = IPC_NOWAIT;
	/* test if the semaphore really is set to 0 */
	if (semop(Mainmem.semid, &sembuf, 1) != 0) {
	    break; /* should not happen since waitcnt is non-null */
	}
	sembuf.sem_op = INC_SEM;
	semop(Mainmem.semid, &sembuf, 1);
	/* The server gets an ALRM signal which will prevent this system
	   call from blocking forever if a client has died while waiting */
	sembuf.sem_op = TEST0_SEM;
	sembuf.sem_flg = 0;
	err = 0;
	while (semop(Mainmem.semid, &sembuf, 1) == -1 && errno == EINTR) {
	    ++err;
	    if (err % 1000/MSPTICK == 0) /* One second timeout interval */
		if (kill(usaddr->pid,0) == -1 && errno == ESRCH) {
		    sembuf.sem_op = WAITDEC_SEM; /* reset semaphore to zero */
		    semop(Mainmem.semid, &sembuf, 1);
		    break; 		    /* since the process has exited */
		}
	}
	usaddr->waitcnt--;
	--n;
    }
    /* awaken processes doing a select() */
    sembuf.sem_num = SEM_SELECT;
    sembuf.sem_op = *Selectcnt;
    sembuf.sem_flg = 0;
    semop(Mainmem.semid, &sembuf, 1);
    sembuf.sem_op = TEST0_SEM;
    err = 0;
    while(semop(Mainmem.semid, &sembuf, 1) == -1 && errno == EINTR)
	if(++err == 1000/MSPTICK) { /* one second timeout here */
	    sembuf.sem_op = WAITDEC_SEM;
	    sembuf.sem_flg = IPC_NOWAIT; /* reset the semaphore to zero */
	    while(semop(Mainmem.semid, &sembuf, 1) != -1)
		;
	    break;
	}
    *Selectcnt = 0;
    releaseusock();
}

/* The following two functions are for managing access to the Usock table.
   they are similar to dirps() and restore(), but subsequent calls to
   holdusock() by the same process will block.
*/
holdusock()
{
    struct sembuf sembuf;
    sembuf.sem_num = SEM_USOCK;
    sembuf.sem_flg = 0;
    sembuf.sem_op = WAITDEC_SEM;
    while (semop(Mainmem.semid, &sembuf, 1) == -1 && errno == EINTR)
	;
}

releaseusock()
{
    struct sembuf sembuf;
    sembuf.sem_num = SEM_USOCK;
    sembuf.sem_flg = 0;
    sembuf.sem_op = INC_SEM;
    semop(Mainmem.semid, &sembuf, 1);
}    


/* Disassociate the current proces with the socket refered to by usock and
   return the number of other processes using the socket. */
socketinuse(up)
struct usock *up;
{
    int i,pid,found = 0;
    pid = getpid();
    holdusock();
    for(i=0; i < sizeof(up->pid)/sizeof(int); ++i) {
	if(up->pid[i] == pid)
	    up->pid[i] = 0;
	if(up->pid[i] && kill(up->pid[i],0) == -1 && errno == ESRCH)
	    continue;
	if(up->pid[i])
	    found++;
    }
    releaseusock();
    if(found)
	return found;
    return 0;
}

int32
rpcsend(bp,noret)
struct mbuf *bp;
int noret;
{
    struct mrec *mrec;
    struct sembuf sembuf;
    int32 retval;
    sembuf.sem_op = WAITDEC_SEM;
    sembuf.sem_flg = 0;
    sembuf.sem_num = SEM_RPC;
    /* wait until it is our turn */
    while (semop(Mainmem.semid, &sembuf, 1) == -1 && errno == EINTR)
	;
    *Mainmem.queue = bp;
    sembuf.sem_op = WAITDEC_SEM;
    sembuf.sem_flg = 0;
    sembuf.sem_num = SEM_RPCREPLY;
    /* signal that the command data has been queued */
    while (semop(Mainmem.semid, &sembuf, 1) == -1 && errno == EINTR)
	;
    if(noret)		/* no return value wanted for this command */
	return 0;
    /* wait for the reply to become ready */
    while (semop(Mainmem.semid, &sembuf, 1) == -1 && errno == EINTR)
	;
    mrec = (struct mrec *)Mainmem.shmbase;
    retval = mrec->retval;
    sembuf.sem_op = 2;	/* a "double" INC_SEM */
    /* signal that we are all done */
    semop(Mainmem.semid, &sembuf, 1);
    return retval;
}
#ifdef SYS5
int getdtablesize()
{
#ifdef NOFILES_MAX
	return NOFILES_MAX;
#else
	return NOFILE;
#endif
}
#endif
