#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/errno.h>
#include "global.h"
#include "memory.h"
#include "sems.h"
#include "mbuf.h"
#include "netuser.h"
#include "ax25.h"
#include "nr4.h"
#include "rpc.h"
#include "upcalls.h"

extern struct memory Mainmem;

extern struct rpcent rpctab[];
int16 Lport = 1024;
extern int Axi_sock;
extern struct ax25_addr Mycall;
extern int32 Ip_addr;
grabinput()
{
	struct sembuf sembuf;
	struct mrec *mrec;
	struct mbuf *bp;
	int cmd, lenidx, i, size;
	int32 arg[12];
	char *p;
	mrec = (struct mrec *)Mainmem.shmbase;
	sembuf.sem_num = SEM_RPCREPLY;
	sembuf.sem_op = TEST0_SEM;
	sembuf.sem_flg = 0;
	/* wait until the data becomes available */
	while((i = semop(Mainmem.semid, &sembuf, 1)) == -1 && errno == EINTR)
	    ;
	bp = *Mainmem.queue;
	if(bp == NULLBUF){
	    printf("fatal error, grabinput() got no data\n");
	    memstat();
	}
	mrec->retval = mrec->retval2 = 0;
	pullup(&bp,&cmd,sizeof(int));
/*
printf("cmd %d\n",cmd);
*/
	p = rpctab[cmd].fmt;
	lenidx = 0;
	if (rpctab[cmd].len[0])  /* if there are 'x' records */
	    size = sizeof(int); /* skip the non-NULL flag */
	else size = 0;
	for (i=0; *p; ++i, ++p) {
	    if (*p == 'c') {
		arg[i] = (int32) *(bp->data + size);
		++size;
	    }
	    if (*p == 'i') {
		int tmp;
		memcpy(&tmp,bp->data + size,sizeof(int));
		asy[i] = (int32) tmp;
		size += sizeof(int);
	    }
	    if (*p == '2') {
		int16 tmp;
		memcpy(&tmp,bp->data + size,sizeof(int16));
		asy[i] = (int32) tmp;
		size += sizeof(int16);
	    }
	    if (*p == '4') {
		memcpy(&arg[i],bp->data + size,sizeof(int32));
		size += sizeof(int32);
	    }
	    if (*p == 'p') {
		void *tmp;
		memcpy(&tmp,bp->data + size,sizeof(void *));
		arg[i] = (int32) tmp;
		size += sizeof(p);
	    }
	    if (*p == 'x') {
		int tmp;
		memcpy(&tmp,bp->data,sizeof(int));
		if (tmp & (1 << lenidx))
		    arg[i] = (int32)(bp->data + size);
		else arg[i] = 0;
		size += rpctab[cmd].len[lenidx];
		++lenidx;
	    }
	}
	switch (cmd) {
	case RPC_INIT:
	    mrec->retval = initclimem((key_t)arg[0]);
	    break;
	case RPC_RAWNR:
	    mrec->retval = (int32) raw_nr((char)arg[0]);
	    break;
	case RPC_RAWIP:
	    mrec->retval = (int32) raw_ip((char)arg[0],uptab[arg[1]]);
	    break;
	case RPC_OPENUDP:
	    mrec->retval = (int32) open_udp((struct socket *) arg[0],
					    uptab[arg[1]]);
	    break;
	case RPC_OPENTCP:
	    mrec->retval = (int32) open_tcp((struct socket *) arg[0],
			   (struct socket *) arg[1], (int)arg[2], (int)arg[3],
			   uptab[arg[4]], uptab[arg[5]], uptab[arg[6]],
			   (int)arg[7], (int)arg[8]);
	    break;
	case RPC_OPENAX25:
	    mrec->retval = (int32) open_ax25((struct iface *) arg[0],
			   (struct ax25_addr *) arg[1], (struct ax25_addr *)
			   arg[2], (int)arg[3], (int)arg[4], uptab[arg[5]],
			   uptab[arg[6]], uptab[arg[7]], (int)arg[8]);
	    break;
	case RPC_OPENNR4:
	    mrec->retval = (int32) open_nr4((struct nr4_addr *) arg[0],
			   (struct nr4_addr *)arg[1],(int)arg[2],uptab[arg[3]],
			   uptab[arg[4]], uptab[arg[5]], (int)arg[6]);
	    break;
	case RPC_IFLOOKUP:
	    mrec->retval = (int32) if_lookup((char *)arg[0]);
	    break;
	case RPC_FINDAX25:
	    mrec->retval = (int32) find_ax25((struct ax25_addr *)arg[0]);
	    break;
	case RPC_RECVTCP:
	    mrec->retval = (int32) recv_tcp((struct tcb *) arg[0],
				    (struct mbuf **)arg[1],(int)arg[2]);
	    break;
	case RPC_RECVUDP:
	    mrec->retval = (int32) recv_udp((struct udp_cb *) arg[0],
			    (struct socket *)arg[1], (struct mbuf **)arg[2]);
	    break;
	case RPC_RECVAX25:
	    mrec->retval = (int32) recv_ax25((struct ax25_cb *) arg[0],
					     (int)arg[1]);
	    break;
	case RPC_RECVNR4:
	    mrec->retval = (int32)recv_nr4((struct nr4cb *)arg[0],(int)arg[1]);
	    break;
	case RPC_SENDTCP:
	    mrec->retval = (int32) send_tcp((struct tcb *) arg[0],(struct mbuf *)arg[1]);
	    break;
	case RPC_SENDUDP:
	    mrec->retval = (int32) send_udp((struct socket *) arg[0],
			    (struct socket *)arg[1],(char)arg[2],(char)arg[3],
			    (struct mbuf *)arg[4],(int16)arg[5],(int16)arg[6],
			    (char)arg[7]);
	    break;
	case RPC_IPSEND:
	    mrec->retval = (int32) ip_send(arg[0],arg[1],(char)arg[2],
			    (char)arg[3],(char)arg[4],(struct mbuf *)arg[5],
			    (int16)arg[6],(int16)arg[7],(char)arg[8]);
	    break;
	case RPC_SENDAX25:
	    mrec->retval = (int32) send_ax25((struct ax25_cb *) arg[0],
					   (struct mbuf *)arg[1],(int)arg[2]);
	    break;
	case RPC_AXOUTPUT:
	    mrec->retval = (int32) ax_output((struct iface *) arg[0],
			    (char *)arg[1],(char *)arg[2],(int16)arg[3],
			    (struct mbuf *)arg[4]);
	    break;
	case RPC_NRSENDRAW:
	    mrec->retval = (int32) nr_sendraw((struct ax25_addr *) arg[0],
			    (int)arg[1],(int)arg[2],(struct mbuf *)arg[3]);
	    break;
	case RPC_SENDNR4:
	    mrec->retval = (int32) send_nr4((struct nr4cb *) arg[0],
					    (struct mbuf *) arg[1]);
	    break;
	case RPC_KICKTCP:
	    mrec->retval = (int32) kick_tcp((struct tcb *) arg[0]);
	    break;
	case RPC_KICKAX25:
	    mrec->retval = (int32) kick_ax25((struct ax25_cb *) arg[0]);
	    break;
	case RPC_KICKNR4:
	    mrec->retval = (int32) kick_nr4((struct nr4cb *) arg[0]);
	    break;
	case RPC_CLOSETCP:
	    mrec->retval = (int32) close_tcp((struct tcb *) arg[0]);
	    break;
	case RPC_RESETTCP:
	    reset_tcp((struct tcb *) arg[0]);
	    break;
	case RPC_DISCAX25:
	    mrec->retval = (int32) disc_ax25((struct ax25_cb *) arg[0]);
	    break;
	case RPC_RESETAX25:
	    mrec->retval = (int32) reset_ax25((struct ax25_cb *) arg[0]);
	    break;
	case RPC_DISCNR4:
	    (void) disc_nr4((struct nr4cb *) arg[0]);
	    break;
	case RPC_RESETNR4:
	    (void) reset_nr4((struct nr4cb *) arg[0]);
	    break;
	case RPC_DELUDP:
	    mrec->retval = (int32) del_udp((struct udp_cb *) arg[0]);
	    break;
	case RPC_DELRNR:
	    (void) del_rnr((struct raw_nr *)arg[0]);
	    break;
	case RPC_DELIP:
	    (void) del_ip((struct raw_ip *)arg[0]);
	    break;
	case RPC_GETUPC:
	    mrec->retval = (int32) uptab[arg[0]];
	    break;
	case RPC_AXISOCK:
	    if (Axi_sock == -1 || arg[0] == -1)
		Axi_sock == (int) arg[0];
	    mrec->retval = (int32) Axi_sock;
	    break;
	case RPC_GETLPORT:
	    mrec->retval = (int32) Lport++;
	    break;
	case RPC_GETSET:
	    mrec->retval = (int32) shmalloc(sizeof(struct settings));
	    (*(struct settings *)mrec->retval).Ip_addr = Ip_addr;
	    memcpy(&(*(struct settings *)mrec->retval).Mycall,&Mycall,
		   sizeof(struct ax25_addr));
	    break;
	case RPC_PSIGNAL:
	    (void) psignal((struct usock *)arg[0], (int)arg[1]);
	    break;
	case RPC_STOPWAIT:
	    break;
	default:
	    printf("what is this?\n");
	    mrec->retval = -1;
	    break;
	}
	free_p(bp);
	sembuf.sem_num = SEM_RPCREPLY;
	sembuf.sem_flg = 0;
	sembuf.sem_op = INC_SEM;
	/* Signal that the reply is ready */
	semop(Mainmem.semid, &sembuf, 1);
	if(rpctab[cmd].noret) /* the client is not intereseted though */
	    return;
	sembuf.sem_op = -2; 	/* a "double" WAITDEC_SEM */
/*	sembuf.sem_flg = SEM_UNDO;*/
	/* wait until the reply has been received */
	while (semop(Mainmem.semid, &sembuf, 1) == -1 && errno == EINTR)
	    ;
	sembuf.sem_op = INC_SEM;
	semop(Mainmem.semid, &sembuf, 1);
}

