#include "global.h"
#include "mbuf.h"
#include "tcp.h"
#include "ip.h"
#include "netrom.h"
#include "iface.h"
#include "socket.h"
#include "rpc.h"

/* The format of each remote procedure call is specified in rpctab.
   "i" denotes an integer, "c" a character, "2" a 16 bit integer,
   and "4" a 32 bit integer, all passed by value. "p" is for a pointer
   to something in the Mainmem shared memory segment. "x" denotes a
   pointer to an element that is not in the Mainmem shared memory segment.
   The elemet will be copied into shared memory, so any changes to it by
   the remote function will be lost. The length of that element should be
   specified in the first unused integer in the 'len' array.
*/
struct rpcent rpctab[] = {
    {""},
    {"i"},					/* initownmem */
    {"c"},					/* raw_nr */
    {"ci"},					/* raw_ip */
    {"xi",0,sizeof(struct socket)},		/* open_udp */
    {"xxiiiiiii",0,sizeof(struct socket),sizeof(struct socket)}, /* open_tcp */
    {"pxxiiiiii",0,sizeof(struct ax25_addr),sizeof(struct ax25_addr)},
						/* open_ax25 */
    {"xxiiiii",0,sizeof(struct nr4_addr),sizeof(struct nr4_addr)},/* open_nr4*/
    {"x",0,ILEN},					/* if_lookup */
    {"x",0,sizeof(struct ax25_addr)},		/* find_ax25 */
    {"ppi",0},					/* recv_tcp */
    {"pxp",0,sizeof(struct socket)},		/* recv_udp */
    {"pi"},					/* recv_ax25 */
    {"pi"},					/* recv_nr4 */
    {"pp",1},					/* send_tcp */
    {"xxccp22c",0,sizeof(struct socket),sizeof(struct socket)}, /* send_udp */
    {"44cccp22c"},				/* ip_send */
    {"ppi"},					/* send_ax25 */
    {"pxx2p",0,AXALEN,AXALEN},			/* ax_output */
    {"xiip",0,sizeof(struct ax25_addr)},	/* nr_sendraw */
    {"pp"},					/* send_nr4 */
    {"p"},					/* kick_tcp */
    {"p"},					/* kick_ax25 */
    {"p"},					/* kick_nr4 */
    {"p"},					/* close_tcp */
    {"p"},					/* reset_tcp */
    {"p"},					/* disc_ax25 */
    {"p"},					/* reset_ax25 */
    {"p",1},					/* disc_nr4 */
    {"p",1},					/* reset_nr4 */
    {"p"},					/* del_udp */
    {"p",1},					/* del_rnr */
    {"p",1},					/* del_ip */
    {"i"},					/* "get_upc" */
    {"i"},					/* "set_Axi_sock" */
    {""},					/* "get_Lport" */
    {""},					/* "get_settings" */
    {"pi",1},					/* psignal */
    {"",1}					/* "stop and wait" */
};
