/* Send and receive IP datagrams on serial lines. Compatible with SLIP
 * under Berkeley Unix.
 */
#include <stdio.h>
#include "global.h"
#include "mbuf.h"
#include "iface.h"
#include "ax25.h"
#include "slip.h"
#include "asy.h"
#include "trace.h"

int kiss_ioctl(),slip_send();
int16 asy_recv();
static int slipq();
static void asy_start();
static struct mbuf *slip_encode();

/* Slip level control structure */
struct slip Slip[ASY_MAX];

/* Send routine for point-to-point slip
 * This is a trivial function since there is no slip link-level header
 */
int
slip_send(data,iface,gateway,prec,del,tput,rel)
struct mbuf *data;	/* Buffer to send */
struct iface *iface;	/* Pointer to interface control block */
int32 gateway;		/* Ignored (SLIP is point-to-point) */
int prec;
int del;
int tput;
int rel;
{
	if(iface == NULLIF){
		free_p(data);
		return -1;
	}
	return (*iface->raw)(iface,data);
}
/* Send a raw slip frame -- also trivial */
int
slip_raw(iface,data)
struct iface *iface;
struct mbuf *data;
{
	dump(iface,IF_TRACE_OUT,Slip[iface->dev].type,data);
	/* Queue a frame on the slip output queue and start transmitter */
	return slipq(iface->dev,data);
}
/* Encode a raw packet in slip framing, put on link output queue, and kick
 * transmitter
 */
static int
slipq(dev,data)
int16 dev;		/* Serial line number */
struct mbuf *data;	/* Buffer to be sent */
{
	register struct slip *sp;
	struct mbuf *bp;

	if((bp = slip_encode(data)) == NULLBUF)
		return -1;	

	sp = &Slip[dev];
	enqueue(&sp->sndq,bp);
	sp->sndcnt++;
	asy_start(dev);
	return 0;
}
/* Start output, if possible, on asynch device dev */
static void
asy_start(dev)
int16 dev;
{
	register struct slip *sp;
	char i_state;

	i_state = dirps();
	if(!stxrdy(dev)){
		restore(i_state);
		return;		/* Transmitter not ready */
	}
	sp = &Slip[dev];
	if(sp->tbp != NULLBUF){
		/* transmission just completed */
		free_p(sp->tbp);
		sp->tbp = NULLBUF;
	}
	if(sp->sndq == NULLBUF){
		restore(i_state);
		return;	/* No work */
	}
	sp->tbp = dequeue(&sp->sndq);
	sp->sndcnt--;
	asy_output(dev,sp->tbp->data,sp->tbp->cnt);
	restore(i_state);
}
/* Encode a packet in SLIP format */
static
struct mbuf *
slip_encode(bp)
struct mbuf *bp;
{
	struct mbuf *lbp;	/* Mbuf containing line-ready packet */
	register char *cp;
	char c;

	/* Allocate output mbuf that's twice as long as the packet.
	 * This is a worst-case guess (consider a packet full of FR_ENDs!)
	 */
	lbp = alloc_mbuf(2*len_mbuf(bp) + 2);
	if(lbp == NULLBUF){
		/* No space; drop */
		free_p(bp);
		return NULLBUF;
	}
	cp = lbp->data;

	/* Flush out any line garbage */
	*cp++ = FR_END;

	/* Copy input to output, escaping special characters */
	while(pullup(&bp,&c,1) == 1){
		switch(uchar(c)){
		case FR_ESC:
			*cp++ = FR_ESC;
			*cp++ = T_FR_ESC;
			break;
		case FR_END:
			*cp++ = FR_ESC;
			*cp++ = T_FR_END;
			break;
		default:
			*cp++ = c;
		}
	}
	*cp++ = FR_END;
	lbp->cnt = cp - lbp->data;
	return lbp;
}
void
asytxdone(dev)
int16 dev;
{
	/* Kick the transmitter if it's idle */
	if(stxrdy(dev))
		asy_start(dev);
}
