#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/ioctl.h>
#include <stropts.h>
#ifdef	__STDC__
#include <stdarg.h>
#endif
#include "global.h"
#include "mbuf.h"
#include "netuser.h"
#include "iface.h"
#include "ip.h"
#include "tcp.h"
#include "udp.h"
#include "ax25.h"
#include "lapb.h"
#include "netrom.h"
#include "nr4.h"
#include "usock.h"
#include "socket.h"
#include "config.h"
#include "rpc.h"
#include "sems.h"
#include "memory.h"
#include "upcalls.h"
    
int Initok = 0;
extern char *Tcpstates[],*Tcpreasons[],*Axreasons[];
extern char *Ax25states[];
extern char *pinet();
static int checkaddr();
static void autobind();
static void (*trdiscard)() = NULLVFP; /* will be filled in by a rpcall */
static int Alarmexp;
struct ax25_addr Mycall;
int32 Ip_addr;
extern struct usock *itop();
extern struct memory Mainmem;

char *Socktypes[] = {
	"Not Used",
	"TCP",
	"UDP",
	"AX25 I",
	"AX25 UI",
	"Raw IP",
	"NETROM3",
	"NETROM"
};
char  Badsocket[] = "Bad socket";
struct usock *Usock;		/* Socket entry array */
int Nusock = DEFNSOCK;		/* Number of socket entries */

#ifdef	AX25
static int Axi_sock = -1;	/* Socket number listening for AX25 connections */
int Axui_sock = -1;	/* Socket number listening for AX25 UI frames */
struct mbuf *Bcq;	/* Queue of incoming UI frames */
extern int16 Axwindow;	/* Window size in bytes */
#endif

/* Create a user socket, return socket index
 * The mapping to actual protocols is as follows:
 *		
 *		
 * ADDRESS FAMILY	Stream		Datagram	Raw	    Seq. Packet
 *
 * AF_INET		TCP		UDP		IP
 * AF_AX25		I-frames	UI-frames
 * AF_NETROM						NET/ROM L3  NET/ROM L4
 */
int
socket(af,type,protocol)
int af;		/* Address family */
int type;	/* Stream or datagram */
int protocol;	/* Used for raw IP sockets */
{
	register struct usock *up;
	int s,i;

	if (!Initok)
	    shminit();
	holdusock();
	for(up=Usock;up < &Usock[Nusock];up++)
		if(up->type == NOTUSED)
			break;

	if(up == &Usock[Nusock]){
		/* None left */
		releaseusock();
		errno = EMFILE;
		return -1;
	}
	s = up - Usock + USOCKBASE;
	errno = 0;
	up->pid[0] = getpid();
	for(i=1; i < sizeof(up->pid)/sizeof(int); ++i)
	    up->pid[i] = 0;
	up->waitcnt = 0;
	up->sigpid = 0;
	up->noblock = 0;
	up->rdysock = -1;
	up->cb.p = NULLCHAR;
	up->peername = up->name = NULLCHAR;
	up->namelen = up->peernamelen = 0;
	memset(up->errcodes,0,sizeof(up->errcodes));
	switch(af){
#ifdef	AX25
	case AF_AX25:
		switch(type){
		case SOCK_STREAM:
			up->type = TYPE_AX25I;
			break;
		case SOCK_DGRAM:
			up->type = TYPE_AX25UI;
			break;
		default:
			errno = ESOCKTNOSUPPORT;
			break;
		}
		break;
#endif
#ifdef NETROM
	case AF_NETROM:
		switch(type){
		case SOCK_RAW:
			up->type = TYPE_NETROML3;
			up->cb.rnr=(struct raw_nr *)rpcall(RPC_RAWNR,protocol);
			break;
		case SOCK_SEQPACKET:
			up->type = TYPE_NETROML4;
			break;
		default:
			errno = ESOCKTNOSUPPORT;
			break;
		}
		break;
#endif
	case AF_INET:
		switch(type){
		case SOCK_STREAM:
			up->type = TYPE_TCP;
			break;
		case SOCK_DGRAM:
			up->type = TYPE_UDP;
			break;
		case SOCK_RAW:
			up->type = TYPE_RAW;
			up->cb.rip = (struct raw_ip *)
			    rpcall(RPC_RAWIP,protocol,UP_RIPRECV);
                        up->cb.rip->user = s;
			break;
		default:
			errno = ESOCKTNOSUPPORT;
			break;
		}
		break;
	default:
		errno = EAFNOSUPPORT;
		break;
	}
	releaseusock();
	if(errno)
		return -1;

	return s;
}

/* Attach a local address/port to a socket. If not issued before a connect
 * or listen, will be issued automatically
 */
int
bind(s,name,namelen)
int s;		/* Socket index */
char *name;	/* Local name */
int namelen;	/* Length of name */
{
	register struct usock *up;
	union sp local;
	struct socket lsock;

	if (!Initok)
	    shminit();
	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(name == NULLCHAR){
		errno = EFAULT;
		return -1;
	}
	if(up->name != NULLCHAR){
		/* Bind has already been issued */
		errno = EINVAL;
		return -1;
	}
	if(checkaddr(up->type,name,namelen) == -1){
		/* Incorrect length or family for chosen protocol */
		errno = EAFNOSUPP;
		return -1;	
	}
	/* Stash name in an allocated block */
	up->namelen = namelen;
	up->name = shmalloc(namelen);
	memcpy(up->name,name,namelen);
	/* Create control block for datagram sockets */
	switch(up->type){
	case TYPE_UDP:
		local.in = (struct sockaddr_in *)up->name;
		lsock.address = local.in->sin_addr.s_addr;
		lsock.port = local.in->sin_port;
		up->cb.udp = (struct udp_cb *) rpcall(RPC_OPENUDP,&lsock,
						      UP_SURCALL);
		up->cb.udp->user = s;
		break;
#ifdef	AX25
	case TYPE_AX25UI:
		if(Axui_sock != -1){
			errno = EADDRINUSE;
			return -1;
		}
		Axui_sock = s;
		break;
#endif
	}
	return 0;
}
/* Post a listen on a socket */
int
listen(s,backlog)
int s;		/* Socket index */
int backlog;	/* 0 for a single connection, 1 for multiple connections */
{
	register struct usock *up;
	union sp local;
	struct socket lsock;

	if (!Initok)
	    shminit();
	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(up->cb.p != NULLCHAR){
		errno = EISCONN;
		return -1;
	}
	switch(up->type){
	case TYPE_TCP:
		if(up->name == NULLCHAR)
			autobind(s,AF_INET);

		local.in = (struct sockaddr_in *)up->name;
		lsock.address = local.in->sin_addr.s_addr;
		lsock.port = local.in->sin_port;
		up->cb.tcb = (struct tcb *)rpcall(RPC_OPENTCP,&lsock,NULLSOCK,
		 backlog ? TCP_SERVER:TCP_PASSIVE,0,
		UP_STRCALL,UP_STTCALL,UP_STSCALL,0,s);
		break;
#ifdef AX25
	case TYPE_AX25I:
		if(up->name == NULLCHAR)
			autobind(s,AF_AX25);
		if(rpcall(RPC_AXISOCK,s) == s)
			Axi_sock = s;
		else {
			errno = EOPNOTSUPP;
			return -1;
		}
		local.ax = (struct sockaddr_ax *)up->name;
		up->cb.ax25 = (struct ax25_cb *) rpcall(RPC_OPENAX25,NULLIF,
			      &local.ax->ax25_addr,NULLAXADDR,
			      backlog ? AX_SERVER:AX_PASSIVE,0,
			      UP_SARCALL,UP_SATCALL,UP_SASCALL,s);
		break;
#endif
#ifdef NETROM
	case TYPE_NETROML4:
		if(up->name == NULLCHAR)
			autobind(s,AF_NETROM);
		local.nr = (struct sockaddr_nr *)up->name;
		up->cb.nr4 = (struct nr4cb *)rpcall(RPC_OPENNR4,
			     &local.nr->nr_addr,NULLNRADDR,
			     backlog ? AX_SERVER:AX_PASSIVE,
			     UP_SNRCALL,UP_SNTCALL,UP_SNSCALL,s);
		break;
#endif
	default:
		/* Listen not supported on datagram sockets */
		errno = EOPNOTSUPP;
		return -1;		
	}
	return 0;
}
/* Initiate active open. For datagram sockets, merely bind the remote address. */
int
connect(s,peername,peernamelen)
int s;			/* Socket index */
char *peername;		/* Peer name */
int peernamelen;	/* Length of peer name */
{
	register struct usock *up;
	union cb cb;
	union sp local,remote;
	struct socket lsock,fsock;
	struct iface *iface;

	if (!Initok)
	    shminit();
	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(peername == NULLCHAR){
		/* Connect must specify a remote address */
		errno = EFAULT;
		return -1;
	}
	if(checkaddr(up->type,peername,peernamelen) == -1){
		errno = EAFNOSUPPORT;
		return -1;
	}
	/* Raw socket control blocks are created in socket() */
	if(up->type != TYPE_RAW && up->type != TYPE_NETROML3 &&
	   up->cb.p != NULLCHAR){
		errno = EISCONN;
		return -1;
	}
	if((up->peername = shmalloc(peernamelen)) == NULLCHAR){
		errno = ENOMEM;
		return -1;
	}
	memcpy(up->peername,peername,peernamelen);
	up->peernamelen = peernamelen;

	/* Set up the local socket structures */
	if(up->name == NULLCHAR){
		switch(up->type){
		case TYPE_TCP:
		case TYPE_UDP:
		case TYPE_RAW:
			autobind(s,AF_INET);
			break;
		case TYPE_AX25I:
		case TYPE_AX25UI:
			autobind(s,AF_AX25);
			break;
		case TYPE_NETROML3:
		case TYPE_NETROML4:
			autobind(s,AF_NETROM);
			break;
		}
	}
	switch(up->type){
	case TYPE_TCP:
		/* Construct the TCP-style ports from the sockaddr structs */
		local.in = (struct sockaddr_in *)up->name;
		lsock.address = local.in->sin_addr.s_addr;
		lsock.port = local.in->sin_port;

		remote.in = (struct sockaddr_in *)up->peername;
		fsock.address = remote.in->sin_addr.s_addr;
		fsock.port = remote.in->sin_port;

		/* Open the TCB in active mode */
		up->cb.tcb = (struct tcb *)rpcall(RPC_OPENTCP,&lsock,&fsock,
			TCP_ACTIVE,0,UP_STRCALL,UP_STTCALL,UP_STSCALL,0,s);

		/* Wait for the connection to complete */
		while((cb.tcb = up->cb.tcb) != NULLTCB && cb.tcb->state != ESTABLISHED){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.tcb == NULLTCB){
			/* Probably got refused */
			shfree(up->peername);
			up->peername = NULLCHAR;
			errno = ECONNREFUSED;
			return -1;
		}
		break;
	case TYPE_UDP:
	case TYPE_AX25UI:
	case TYPE_NETROML3:
	case TYPE_RAW:
		/* Control block already created by bind() */
		break;
#ifdef	AX25
	case TYPE_AX25I:
		local.ax = (struct sockaddr_ax *)up->name;
		remote.ax = (struct sockaddr_ax *)up->peername;
		if((iface = (struct iface *)rpcall(RPC_IFLOOKUP,
			    remote.ax->iface)) == NULLIF){
			errno = EINVAL;
			return -1;
		}
		/* If we already have an AX25 link we can use it */
		if((up->cb.ax25 = (struct ax25_cb *) rpcall(RPC_FINDAX25,
		   &remote.ax->ax25_addr)) != NULLAX25
		   && up->cb.ax25->state != DISCONNECTED &&
		   up->cb.ax25->user == -1) {
			up->cb.ax25->user = s;
			up->cb.ax25->r_upcall = (void(*)())rpcall(RPC_GETUPC,
								UP_SARCALL);
			up->cb.ax25->t_upcall = (void(*)())rpcall(RPC_GETUPC,
								UP_SATCALL);
			up->cb.ax25->s_upcall = (void(*)())rpcall(RPC_GETUPC,
								UP_SASCALL);
			if(up->cb.ax25->state == CONNECTED
			   || up->cb.ax25->state == RECOVERY)
			    	return 0;
		}
		else up->cb.ax25 = (struct ax25_cb *) rpcall(RPC_OPENAX25,
				   iface,&local.ax->ax25_addr,
				   &remote.ax->ax25_addr,AX_ACTIVE,Axwindow,
				   UP_SARCALL,UP_SATCALL,UP_SASCALL,s);

		/* Wait for the connection to complete */
		while((cb.ax25 = up->cb.ax25) != NULLAX25 && cb.ax25->state != CONNECTED){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.ax25 == NULLAX25){
			/* Connection probably already exists */
			shfree(up->peername);
			up->peername = NULLCHAR;
			errno = ECONNREFUSED;
			return -1;
		}
		break;
#endif
#ifdef	NETROM
	case TYPE_NETROML4:
		local.nr = (struct sockaddr_nr *)up->name;
		remote.nr = (struct sockaddr_nr *)up->peername;
		up->cb.nr4 = (struct nr4cb *) rpcall(RPC_OPENNR4,
			     &local.nr->nr_addr,&remote.nr->nr_addr,
			     AX_ACTIVE,UP_SNRCALL,UP_SNTCALL,UP_SNSCALL,s);

		/* Wait for the connection to complete */
		while((cb.nr4 = up->cb.nr4) != NULLNR4CB && cb.nr4->state != NR4STCON){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.nr4 == NULLNR4CB){
			/* Connection probably already exists */
			shfree(up->peername);
			up->peername = NULLCHAR;
			errno = ECONNREFUSED;
			return -1;
		}
		break;
#endif
	}
	return 0;
}
/* Wait for a connection. Valid only for connection-oriented sockets. */
int
accept(s,peername,peernamelen)
int s;			/* Socket index */
char *peername;		/* Peer name */
int *peernamelen;	/* Length of peer name */
{
	int i;
	register struct usock *up;

	if (!Initok)
	    shminit();
	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(up->cb.p == NULLCHAR){
		errno = EOPNOTSUPP;
		return -1;
	}
	/* Accept is valid only for stream sockets */
	switch(up->type){
	case TYPE_TCP:
	case TYPE_AX25I:
	case TYPE_NETROML4:
		break;
	default:
		errno = EOPNOTSUPP;
		return -1;
	}	
	/* Wait for the state-change upcall routine to signal us */
	while(up->cb.p != NULLCHAR && up->rdysock == -1){
		if(up->noblock){
			errno = EWOULDBLOCK;
			return -1;
		} else if(pwait(up) != 0){
			errno = EINTR;
			return -1;
		}
	}
	if(up->cb.p == NULLCHAR){
		/* Blown away */
		errno = EBADF;
		return -1;
	}
	i = up->rdysock;
	up->rdysock = -1;

	up = &Usock[i-USOCKBASE];
	if(peername != NULLCHAR && peernamelen != NULL){
		*peernamelen = min(up->peernamelen,*peernamelen);
		memcpy(peername,up->peername,*peernamelen);
	}
	return i;
}
/* Low-level receive routine. Passes mbuf back to user; more efficient than
 * higher-level functions recv() and recvfrom(). Datagram sockets ignore
 * the len parameter.
 */
int
recv_mbuf(s,bpp,len,flags,from,fromlen)
int s;			/* Socket index */
struct mbuf **bpp;	/* Place to stash receive buffer */
int len;		/* Maximum length to read (0 implies all) */
int flags;		/* Unused; will control out-of-band data, etc */
char *from;		/* Peer address (only for datagrams) */
int *fromlen;		/* Length of peer address */
{
	register struct usock *up;
	int cnt;
	union cb cb;
	struct socket fsocket;
	struct nr3hdr n3hdr;
	union sp remote;
	struct ip ip;
	struct mbuf **bpq;

	if (!Initok)
	    shminit();
	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	bpq = (struct mbuf **) shcalloc(1,sizeof(*bpq));
	switch(up->type){
	case TYPE_TCP:
		if(trdiscard == NULLVFP) /* this is done only once */
			trdiscard = (void(*)())rpcall(RPC_GETUPC,UP_TRDISCARD);
		while((cb.tcb = up->cb.tcb) != NULLTCB
		 && cb.tcb->r_upcall != trdiscard
/*                 && (cnt = client_recvtcp(cb.tcb,bpq,len)) == -1){*/
                 && (cnt = (int32) rpcall(RPC_RECVTCP,cb.tcb,bpq,len)) == -1){
		 	if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.tcb == NULLTCB){
			/* Connection went away */
			errno = ENOTCONN;
			return -1;
		} else if(cb.tcb->r_upcall == trdiscard){
			/* Receive shutdown has been done */
			errno = ENOTCONN;	/* CHANGE */
			return -1;
		}
		break;
	case TYPE_UDP:
		while((cb.udp = up->cb.udp) != NULLUDP
		&& (cnt = (int)rpcall(RPC_RECVUDP,cb.udp,&fsocket,bpq)) == -1){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.udp == NULLUDP){
			/* Connection went away */
			errno = ENOTCONN;
			return -1;
		}
		if(from != NULLCHAR && fromlen != NULLINT && *fromlen >= SOCKSIZE){
			remote.in = (struct sockaddr_in *)from;
			remote.in->sin_family = AF_INET;
			remote.in->sin_addr.s_addr = fsocket.address;
			remote.in->sin_port = fsocket.port;
			*fromlen = SOCKSIZE;
		}
		break;
	case TYPE_RAW:
		while((cb.rip = up->cb.rip) != NULLRIP
		 && cb.rip->rcvq == NULLBUF){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.rip == NULLRIP){
			/* Connection went away */
			errno = ENOTCONN;
			return -1;
		}
		*bpq = dequeue(&cb.rip->rcvq);
		ntohip(&ip,bpq);
		cnt = len_mbuf(*bpq);
		if(from != NULLCHAR && fromlen != NULLINT && *fromlen >= SOCKSIZE){
			remote.in = (struct sockaddr_in *)from;
			remote.in->sin_family = AF_INET;
			remote.in->sin_addr.s_addr = ip.source;
			remote.in->sin_port = 0;
			*fromlen = SOCKSIZE;
		}
		break;
#ifdef	AX25
	case TYPE_AX25I:
		while((cb.ax25 = up->cb.ax25) != NULLAX25
		 && (*bpq = (struct mbuf *) rpcall(RPC_RECVAX25,cb.ax25,len)) == NULLBUF){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.ax25 == NULLAX25){
			/* Connection went away */
			errno = ENOTCONN;
			return -1;
		}
		cnt = (*bpq)->cnt;
		break;
	case TYPE_AX25UI:
		while(s == Axui_sock && Bcq == NULLBUF){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(&Bcq) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(s != Axui_sock){
			errno = ENOTCONN;
			return -1;
		}
		*bpq = dequeue(&Bcq);

		if(from != NULLCHAR && fromlen != NULLINT
		   && *fromlen >= sizeof(struct sockaddr_ax)){
			pullup(bpq,(char *)from,sizeof(struct sockaddr_ax));
			*fromlen = sizeof(struct sockaddr_ax);
		} else {
			pullup(bpq,NULLCHAR,sizeof(struct sockaddr_ax));
		}
		cnt = len_mbuf(*bpq);
		break;
#endif
#ifdef NETROM
	case TYPE_NETROML3:
		while((cb.rnr = up->cb.rnr) != NULLRNR
		 && cb.rnr->rcvq == NULLBUF){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.rnr == NULLRNR){
			/* Connection went away */
			errno = ENOTCONN;
			return -1;
		}
		*bpq = dequeue(&cb.rnr->rcvq);
		ntohnr3(&n3hdr,bpq);
		cnt = len_mbuf(*bpq);
		if(from != NULLCHAR && fromlen != NULLINT
		   && *fromlen >= sizeof(struct sockaddr_nr)){
			remote.nr = (struct sockaddr_nr *)from;
			remote.nr->nr_family = AF_NETROM;
			/* The callsign of the local user is not part of
			   NET/ROM level 3, so that field is not used here */
			memcpy(remote.nr->nr_addr.node_call,&n3hdr.source,
			       AXALEN);
			*fromlen = sizeof(struct sockaddr_nr);
		}
		break;
	case TYPE_NETROML4:
		while((cb.nr4 = up->cb.nr4) != NULLNR4CB
		      && (*bpq = (struct mbuf *)rpcall(RPC_RECVNR4,cb.nr4,len))
		      == NULLBUF){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.nr4 == NULLNR4CB){
			/* Connection went away */
			errno = ENOTCONN;
			return -1;
		}
		cnt = (*bpq)->cnt;
		break;
#endif
	}
	if(bpp != NULLBUFP)
		*bpp = *bpq;
	shfree(bpq);
	return cnt;
}
/* Low level send routine; user supplies mbuf for transmission. More
 * efficient than send() or sendto(), the higher level interfaces.
 * The "to" and "tolen" parameters are ignored on connection-oriented
 * sockets.
 *
 * In case of error, bp is freed so the caller doesn't have to worry about it.
 */
int
send_mbuf(s,bp,flags,to,tolen)
int s;			/* Socket index */
struct mbuf *bp;	/* Buffer to send */
int flags;		/* not currently used */
char *to;		/* Destination, only for datagrams */
int tolen;		/* Length of destination */
{
	register struct usock *up;
	union cb cb;
	union sp local,remote;
	struct socket lsock,fsock;
	int cnt;

	if (!Initok)
	    shminit();
	if((up = itop(s)) == NULLUSOCK){
		free_p(bp);
		errno = EBADF;
		return -1;
	}
	if(up->name == NULLCHAR){
		/* Automatic local name assignment for datagram sockets */
		switch(up->type){
		case TYPE_UDP:
		case TYPE_RAW:
			autobind(s,AF_INET);
			break;
		case TYPE_AX25UI:
			autobind(s,AF_AX25);
			break;
		case TYPE_NETROML3:
		case TYPE_NETROML4:
			autobind(s,AF_NETROM);
			break;
		default:
			free_p(bp);
			errno = ENOTCONN;
			return -1;
		}
	}
	cnt = len_mbuf(bp);
	switch(up->type){
	case TYPE_TCP:
		if((cb.tcb = up->cb.tcb) == NULLTCB){
			free_p(bp);
			errno = ENOTCONN;
			return -1;
		}		
		cnt = client_sendtcp(cb.tcb,bp);

		while((cb.tcb = up->cb.tcb) != NULLTCB &&
		 cb.tcb->sndcnt > cb.tcb->window){
			/* Send queue is full */
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
 		if(cb.tcb == NULLTCB){
			errno = ENOTCONN;
			return -1;
		}
		break;
	case TYPE_UDP:
		local.in = (struct sockaddr_in *)up->name;
		lsock.address = local.in->sin_addr.s_addr;
		lsock.port = local.in->sin_port;
		if(to != NULLCHAR)
			remote.in = (struct sockaddr_in *)to;
		else if(up->peername != NULLCHAR)
			remote.in = (struct sockaddr_in *)up->peername;
		else {
			errno = ENOTCONN;
			return -1;
		}	
		fsock.address = remote.in->sin_addr.s_addr;
		fsock.port = remote.in->sin_port;
		rpcall(RPC_SENDUDP,&lsock,&fsock,0,0,bp,0,0,0);
		break;
	case TYPE_RAW:
		local.in = (struct sockaddr_in *)up->name;
		if(to != NULLCHAR)
			remote.in = (struct sockaddr_in *)to;
		else if(up->peername != NULLCHAR)
			remote.in = (struct sockaddr_in *)up->peername;
		else {
			errno = ENOTCONN;
			return -1;
		}	
		rpcall(RPC_IPSEND,local.in->sin_addr.s_addr,remote.in->sin_addr.s_addr,
		        up->cb.rip->protocol,0,0,bp,0,0,0);
		break;
#ifdef	AX25
	case TYPE_AX25I:
		if((cb.ax25 = up->cb.ax25) == NULLAX25){
			free_p(bp);
			errno = ENOTCONN;
			return -1;
		}
		rpcall(RPC_SENDAX25,cb.ax25,bp,PID_NO_L3);

		while((cb.ax25 = up->cb.ax25) != NULLAX25 &&
		 len_q(cb.ax25->txq) > cb.ax25->window){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.ax25 == NULLAX25){
			errno = EBADF;
			return -1;
		}
		break;
	case TYPE_AX25UI:
		local.ax = (struct sockaddr_ax *)up->name;
		if(to != NULLCHAR)
			remote.ax = (struct sockaddr_ax *)to;
		else if(up->peername != NULLCHAR)
			remote.ax = (struct sockaddr_ax *)up->peername;
		else {
			errno = ENOTCONN;
			return -1;
		}
		rpcall(RPC_AXOUTPUT,(struct iface *)rpcall(RPC_IFLOOKUP,remote.ax->iface),
			  (char *)&remote.ax->ax25_addr,
			  (char *)&local.ax->ax25_addr,PID_NO_L3,bp);
		break;
#endif
#ifdef NETROM
	case TYPE_NETROML3:
		if(len_mbuf(bp) > NR4MAXINFO) {
			free_p(bp);
			errno = EMSGSIZE;
			return -1;
		}
		local.nr = (struct sockaddr_nr *)up->name;
		if(to != NULLCHAR)
			remote.nr = (struct sockaddr_nr *)to;
		else if(up->peername != NULLCHAR)
			remote.nr = (struct sockaddr_nr *)up->peername;
		else {
			errno = ENOTCONN;
			return -1;
		}	
		/* The NETROM username is always ignored in outgoing traffic */
		rpcall(RPC_NRSENDRAW,(struct ax25_addr *)remote.nr->nr_addr.node_call,
			up->cb.rnr->protocol,up->cb.rnr->protocol,bp);
		break;
	case TYPE_NETROML4:
		if((cb.nr4 = up->cb.nr4) == NULLNR4CB) {
			free_p(bp);
			errno = ENOTCONN;
			return -1;
		}
		if(len_mbuf(bp) > NR4MAXINFO){ /* reject big packets */
			free_p(bp);
			errno = EMSGSIZE;
			return -1;
		}
		rpcall(RPC_SENDNR4,cb.nr4,bp);

		while((cb.nr4 = up->cb.nr4) != NULLNR4CB &&
		 cb.nr4->nbuffered >= cb.nr4->window){
			if(up->noblock){
				errno = EWOULDBLOCK;
				return -1;
			} else if(pwait(up) != 0){
				errno = EINTR;
				return -1;
			}
		}
		if(cb.nr4 == NULLNR4CB){
			errno = EBADF;
			return -1;
		}
		break;
#endif
	}	
	return cnt;
}
/* Higher-level receive routine, intended for connection-oriented sockets.
 * Can be used with datagram sockets, although the sender id is lost.
 */
int
recv(s,buf,len,flags)
int s;		/* Socket index */
char *buf;	/* User buffer */
int len;	/* Max length to receive */
int flags;	/* Unused; will eventually select oob data, etc */
{
	struct mbuf *bp;
	int cnt;

	if(len == 0)
		return 0;	/* Otherwise would be interp as "all" */

	cnt = recv_mbuf(s,&bp,len,flags,NULLCHAR,(int *)NULL);
	if(cnt > 0){
		cnt = min(cnt,len);
		pullup(&bp,buf,cnt);
		free_p(bp);
	}
	return cnt;
}
/* Higher level receive routine, intended for datagram sockets. Can also
 * be used for connection-oriented sockets, although from and fromlen are
 * ignored.
 */
int
recvfrom(s,buf,len,flags,from,fromlen)
int s;		/* Socket index */
char *buf;	/* User buffer */
int len;	/* Maximum length */
int flags;	/* Unused; will eventually select oob data, etc */
char *from;	/* Source address, only for datagrams */
int *fromlen;	/* Length of source address */
{
	struct mbuf *bp;
	register int cnt;

	cnt = recv_mbuf(s,&bp,len,flags,from,fromlen);
	if(cnt > 0){
		cnt = min(cnt,len);
		pullup(&bp,buf,cnt);
		free_p(bp);
	}
	return cnt;
}
/* High level send routine */
int
send(s,buf,len,flags)
int s;		/* Socket index */
char *buf;	/* User buffer */
int len;	/* Length of buffer */
int flags;	/* Unused; will eventually select oob data, etc */
{
	register struct mbuf *bp;
	char sock[MAXSOCKSIZE];
	int i = MAXSOCKSIZE;

	if(getpeername(s,sock,&i) == -1)
		return -1;
	bp = qdata(buf,len);
	return send_mbuf(s,bp,flags,sock,i);
}
/* High level send routine, intended for datagram sockets. Can be used on
 * connection-oriented sockets, but "to" and "tolen" are ignored.
 */
int
sendto(s,buf,len,flags,to,tolen)
int s;		/* Socket index */
char *buf;	/* User buffer */
int len;	/* Length of buffer */
int flags;	/* Unused; will eventually select oob data, etc */
char *to;	/* Destination, only for datagrams */
int tolen;	/* Length of destination */
{
	register struct mbuf *bp;

	if (!Initok)
	    shminit();
	bp = qdata(buf,len);
	return send_mbuf(s,bp,flags,to,tolen);
}
/* Return local name passed in an earlier bind() call */
int
getsockname(s,name,namelen)
int s;		/* Socket index */
char *name;	/* Place to stash name */
int *namelen;	/* Length of same */
{
	register struct usock *up;

	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(name == NULLCHAR || namelen == (int *)NULL){
		errno = EFAULT;
		return -1;
	}
	if(up->name == NULLCHAR){
		/* Not bound yet */
		*namelen = 0;
		return 0;
	}
	if(up->name != NULLCHAR){
		*namelen = min(*namelen,up->namelen);
		memcpy(name,up->name,*namelen);
	}
	return 0;
}
/* Get remote name, returning result of earlier connect() call. */
int
getpeername(s,peername,peernamelen)
int s;			/* Socket index */
char *peername;		/* Place to stash name */
int *peernamelen;	/* Length of same */
{
	register struct usock *up;

	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(up->peername == NULLCHAR){
		errno = ENOTCONN;
		return -1;
	}
	if(peername == NULLCHAR || peernamelen == (int *)NULL){
		errno = EFAULT;
		return -1;
	}
	*peernamelen = min(*peernamelen,up->peernamelen);
	memcpy(peername,up->peername,*peernamelen);
	return 0;
}
/* Return length of protocol queue, either send or receive. */
int
socklen(s,rtx)
int s;		/* Socket index */
int rtx;	/* 0 = receive queue, 1 = transmit queue */
{
	register struct usock *up;

	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(up->cb.p == NULLCHAR){
		errno = ENOTCONN;
		return -1;
	}
	if(rtx < 0 || rtx > 1){
		errno = EINVAL;
		return -1;
	}
	switch(up->type){
	case TYPE_TCP:
		switch(rtx){
		case 0:
			return up->cb.tcb->rcvcnt;
		case 1:
			return up->cb.tcb->sndcnt;
		}
	case TYPE_UDP:
		switch(rtx){
		case 0:
			return up->cb.udp->rcvcnt;
		case 1:
			return 0;
		}
#ifdef	AX25
	case TYPE_AX25I:
		switch(rtx){
		case 0:
			return len_mbuf(up->cb.ax25->rxq);
		case 1:	/* Number of packets, not bytes */
			return len_q(up->cb.ax25->txq);
		}
	case TYPE_AX25UI:
		switch(rtx){	
		case 0:
			return len_q(Bcq);
		case 1:
			return 0;		
		}
#endif
#ifdef NETROM
	case TYPE_NETROML3:
		switch(rtx){	
		case 0:
			return len_q(up->cb.rnr->rcvq);
		case 1:
			return 0;		
		}
	case TYPE_NETROML4:
		switch(rtx){
		case 0:
			return len_mbuf(up->cb.nr4->rxq);
		case 1:	/* Number of packets, not bytes */
			return len_q(up->cb.nr4->txq);
		}
#endif
	case TYPE_RAW:
		switch(rtx){	
		case 0:
			return len_q(up->cb.rip->rcvq);
		case 1:
			return 0;		
		}
	}
	/*NOTREACHED*/
}
/* Force retransmission. Valid only for connection-oriented sockets. */
int
sockkick(s)
int s;	/* Socket index */
{
	register struct usock *up;

	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(up->cb.p == NULLCHAR){
		errno = ENOTCONN;
		return -1;
	}
	switch(up->type){
	case TYPE_TCP:
		rpcall(RPC_KICKTCP,up->cb.tcb);
		break;
#ifdef	AX25
	case TYPE_AX25I:
		rpcall(RPC_KICKAX25,up->cb.ax25);
		break;
#endif
#ifdef NETROM
	case TYPE_NETROML4:
		rpcall(RPC_KICKNR4,up->cb.nr4);
		break;
#endif
	default:
		/* Datagram sockets can't be kicked */
		errno = EOPNOTSUPP;
		return -1;
	}
	return 0;
}

/* Get state of protocol. Valid only for connection-oriented sockets. */
char *
sockstate(s)
int s;		/* Socket index */
{
	register struct usock *up;

	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return NULLCHAR;
	}
	if(up->cb.p == NULLCHAR){
		errno = ENOTCONN;
		return NULLCHAR;
	}
	switch(up->type){
	case TYPE_TCP:
		return Tcpstates[up->cb.tcb->state];
#ifdef	AX25
	case TYPE_AX25I:
		return Ax25states[up->cb.ax25->state];
#endif
#ifdef	NETROM
	case TYPE_NETROML4:
		return Nr4states[up->cb.nr4->state];
#endif
	default:
		/* Datagram sockets don't have state */
		errno = EOPNOTSUPP;
		return NULLCHAR;
	}
	/*NOTREACHED*/
}
/* Close down a socket three ways. Type 0 means "no more receives"; this
 * replaces the incoming data upcall with a routine that discards further
 * data. Type 1 means "no more sends", and obviously corresponds to sending
 * a TCP FIN. Type 2 means "no more receives or sends". This I interpret
 * as "abort the connection".
 */
int
shutdown(s,how)
int s;		/* Socket index */
int how;	/* (see above) */
{
	register struct usock *up;
	
	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(up->cb.p == NULLCHAR){
		errno = ENOTCONN;
		return -1;
	    }
	if(socketinuse(up))
	    return 0;
	switch(up->type){
	case TYPE_TCP:
		switch(how){
		case 0:	/* No more receives -- replace upcall */
			if(trdiscard == NULLVFP) /* this is done only once */
			    trdiscard = (void(*)()) rpcall(RPC_GETUPC,UP_TRDISCARD);
			up->cb.tcb->r_upcall = trdiscard;
			break;
		case 1:	/* Send EOF */
			rpcall(RPC_CLOSETCP,up->cb.tcb);
			break;
		case 2:	/* Blow away TCB */
			rpcall(RPC_RESETTCP,up->cb.tcb);
			up->cb.tcb = NULLTCB;
			break;
		}
		break;
#ifdef	AX25
	case TYPE_AX25UI:
		close_s(s);
		break;
	case TYPE_AX25I:
		switch(how){
		case 0:
		case 1:	/* Attempt regular disconnect */
			rpcall(RPC_DISCAX25,up->cb.ax25);
			break;
		case 2: /* Blow it away */
			rpcall(RPC_RESETAX25,up->cb.ax25);
			up->cb.ax25 = NULLAX25;
			break;
		}
		break;		
#endif
#ifdef	NETROM
	case TYPE_NETROML3:
		close_s(s);
		break;
	case TYPE_NETROML4:
		switch(how){
		case 0:
		case 1:	/* Attempt regular disconnect */
			rpcall(RPC_DISCNR4,up->cb.nr4);
			break;
		case 2: /* Blow it away */
			rpcall(RPC_RESETNR4,up->cb.nr4);
			up->cb.nr4 = NULLNR4CB;
			break;
		}
		break;
#endif
	case TYPE_RAW:
		close_s(s);
		break;
	default:
		errno = EOPNOTSUPP;
		return -1;
	}
	rpcall(RPC_PSIGNAL,up,0); /* we can't do a psignal ourselves */
	return 0;
}
/* Close a socket, freeing it for reuse. Try to do a graceful close on a
 * TCP socket, if possible
 */
int
close_s(s)
int s;		/* Socket index */
{
	register struct usock *up;

	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(socketinuse(up))
		return 0;
	switch(up->type){
	case TYPE_TCP:
		if(up->cb.tcb != NULLTCB){	/* In case it's been reset */
			if(trdiscard == NULLVFP) /* this is done only once */
				trdiscard = (void(*)()) rpcall(RPC_GETUPC,
							    UP_TRDISCARD);
			up->cb.tcb->r_upcall = trdiscard;
			/* Tell the CLOSED upcall there's no more socket */
			up->cb.tcb->user = -1;
			rpcall(RPC_CLOSETCP,up->cb.tcb);
		}
		break;
	case TYPE_UDP:
		if(up->cb.udp != NULLUDP){
			rpcall(RPC_DELUDP,up->cb.udp);
		}
		break;
#ifdef	AX25
	case TYPE_AX25I:
		if(up->cb.ax25 != NULLAX25){
			if(s == Axi_sock){
				rpcall(RPC_AXISOCK,-1);
				Axi_sock = -1;
			}
			/* Tell the CLOSED upcall there's no more socket */
			up->cb.ax25->user = -1;
			rpcall(RPC_DISCAX25,up->cb.ax25);
		}
		break;
	case TYPE_AX25UI:
		Axui_sock = -1;
		free_q(&Bcq);
		rpcall(RPC_PSIGNAL,&Bcq,0);	/* Unblock any reads */
		break;
#endif
#ifdef	NETROM
	case TYPE_NETROML3:
		rpcall(RPC_DELRNR,up->cb.rnr);
		break;
	case TYPE_NETROML4:
		if(up->cb.nr4 != NULLNR4CB){
			/* Tell the CLOSED upcall there's no more socket */
			up->cb.nr4->user = -1;
			rpcall(RPC_DISCNR4,up->cb.nr4);
		}
		break;
#endif
	case TYPE_RAW:
		rpcall(RPC_DELIP,up->cb.rip);
		break;
	default:
		errno = EOPNOTSUPP;
		return -1;
	}
	shfree(up->name);
	shfree(up->peername);

	up->cb.p = NULLCHAR;
	up->name = up->peername = NULLCHAR;
	up->type = NOTUSED;
	rpcall(RPC_PSIGNAL,up,0);/* Wake up anybody doing an accept() or recv() */
	return 0;
}
/* Return ASCII string giving reason for connection closing */
char *
sockerr(s)
int s;	/* Socket index */
{
	register struct usock *up;

	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return Badsocket;
	}
	switch(up->type){
#ifdef	AX25
	case TYPE_AX25I:
		if(up->cb.ax25 != NULLAX25)
			return NULLCHAR;	/* nothing wrong */
		return Axreasons[up->errcodes[0]];
#endif
#ifdef	NETROM
	case TYPE_NETROML4:
		if(up->cb.nr4 != NULLNR4CB)
			return NULLCHAR;	/* nothing wrong */
		return Nr4reasons[up->errcodes[0]];
#endif
	case TYPE_TCP:
		if(up->cb.tcb != NULLTCB)
			return NULLCHAR;	/* nothing wrong */		
		return Tcpreasons[up->errcodes[0]];
	default:
		errno = EOPNOTSUPP;	/* not yet, anyway */
		return NULLCHAR;
	}
}
/* End of user socket primitives, begin user utility subroutines */

/* Receive a newline-terminated line from a socket, returning # chars read */
int
recvline(s,buf,len)
int s;		/* Socket index */
char *buf;	/* User buffer */
unsigned len;	/* Length of buffer */
{
	char c;
	int cnt = 0;
	struct usock *up;

	up = itop(s);
	while(len-- > 1){
		if(recv(s,&c,1,0) != 1){
			*buf = '\0';
			return -1;
		}
		if((up->type == TYPE_AX25I || up->type == TYPE_NETROML4)
		   && uchar(c) == '\r')
			c = '\n';
		*buf++ = c;
		cnt++;
		if(uchar(c) == '\n')
			break;
	}
	if(len != 0)
		*buf = '\0';
	return cnt;
}
/* Do printf on a user socket */
#if	defined(__STDC__) || defined(__TURBOC__)
int
usprintf(int s,char *fmt,...)
{
	register struct mbuf *bp;
	int len;
	va_list args;

	va_start(args,fmt);
	bp = alloc_mbuf(200);
	len = vsprintf(bp->data,fmt,args);
	bp->cnt = strlen(bp->data);
	send_mbuf(s,bp,0,NULLCHAR,0);
	va_end(args);
	return len;
}
#else
/*VARARGS*/
int
usprintf(s,fmt,arg1,arg2,arg3,arg4,arg5,arg6)
int s;			/* Socket index */
char *fmt;		/* Message format */
int arg1,arg2,arg3;	/* Arguments */
int arg4,arg5,arg6;
{
	register struct mbuf *bp;
	int len;

	bp = alloc_mbuf(200);
	len = sprintf(bp->data,fmt,arg1,arg2,arg3,arg4,arg5,arg6);
	bp->cnt = strlen(bp->data);
	send_mbuf(s,bp,0,NULLCHAR,0);
	return len;
}
#endif
/* Convert a socket (address + port) to an ascii string of the form
 * aaa.aaa.aaa.aaa:ppppp
 */
char *
psocket(p)
char *p;	/* Pointer to structure to decode */
{
	char tmp[11];
	static char buf[30];
	union sp sp;
	struct socket socket;

	sp.p = p;
	switch(sp.sa->sa_family){
	case AF_INET:
		socket.address = (int32) sp.in->sin_addr.s_addr;
		socket.port = sp.in->sin_port;
		strcpy(buf,pinet(&socket));
		break;
#ifdef	AX25
	case AF_AX25:
		pax25(tmp,&sp.ax->ax25_addr);
		sprintf(buf,"%s on %s",tmp,sp.ax->iface);
		break;
#endif
#ifdef	NETROM
	case AF_NETROM:
		pax25(tmp,(struct ax25_addr *)sp.nr->nr_addr.user_call);
		sprintf(buf,"%s @ ",tmp);
		pax25(tmp,(struct ax25_addr *)sp.nr->nr_addr.node_call);
		strcat(buf,tmp);
		break;
#endif
	}
	return buf;
}
char *
pinet(s)
struct socket *s;
{
	static char buf[30];

	sprintf(buf,"%s:%u",inet_ntoa(s->address),s->port);
	return buf;
}
/* Read a character from a socket with stream buffering */
int
recvchar(s,bpp)
int s;			/* Socket index */
struct mbuf **bpp;	/* User-supplied buffer pointer */
{
	char c;

	if(*bpp == NULLBUF)
		recv_mbuf(s,bpp,0,0,NULLCHAR,0);	/* Replenish */

	if(pullup(bpp,&c,1) != 1)
		return -1;
	return uchar(c);
}
/* Blow away all sockets belonging to a certain process. Used by killproc(). */
void
freesock(pid)
int pid;
{
	register struct usock *up;
	register int i,j;

	for(i=0;i < Nusock;i++){
		up = &Usock[i];
		if(up->type != NOTUSED)
		    for(j=0; j < sizeof(up->pid)/sizeof(int); ++j)
			if(up->pid[j] == pid)
			    close_s(j+USOCKBASE);
	}
}

/* End of user utility subroutines, start internal subroutines */

/* Verify address family and length according to the socket type */
static int
checkaddr(type,name,namelen)
char *name;
int type;
int namelen;
{
	register union sp sp;

	sp.p = name;
	/* Verify length and address family according to protocol */
	switch(type){
	case TYPE_TCP:
	case TYPE_UDP:
		if(sp.in->sin_family != AF_INET
		 || namelen != sizeof(struct sockaddr_in))
			return -1;
		break;
	case TYPE_AX25I:
	case TYPE_AX25UI:
		if(sp.ax->sax_family != AF_AX25
		 || namelen != sizeof(struct sockaddr_ax))
			return -1;
		break;
	case TYPE_NETROML3:
	case TYPE_NETROML4:
		if(sp.nr->nr_family != AF_NETROM
		 || namelen != sizeof(struct sockaddr_nr))
			return -1;
		break;
	}
	return 0;
}
/* Convert a socket index to an internal user socket structure pointer */
struct usock *
itop(s)
register int s;	/* Socket index */
{
	register struct usock *up;

	s -= USOCKBASE;

	if(s < 0 || s >= Nusock)
		return NULLUSOCK;

	if (!Initok)
	    shminit();
	up = &Usock[s];

	if(up->type == NOTUSED)
		return NULLUSOCK;
	return up;
}
/* Issue an automatic bind of a local address */
static void
autobind(s,af)
int s,af;
{
	char buf[MAXSOCKSIZE];
	union sp sp;

	sp.p = buf;
	switch(af){
	case AF_INET:
		sp.in->sin_family = AF_INET;
		sp.in->sin_addr.s_addr = Ip_addr;
		sp.in->sin_port = (int16) rpcall(RPC_GETLPORT);
		bind(s,sp.p,sizeof(struct sockaddr_in));
		break;
#ifdef	AX25
	case AF_AX25:
		sp.ax->sax_family = AF_AX25;
		memcpy(sp.ax->ax25_addr.call,Mycall.call,ALEN);
		sp.ax->ax25_addr.ssid = Mycall.ssid;
		bind(s,sp.p,sizeof(struct sockaddr_ax));
		break;
#endif
#ifdef	NETROM
	case AF_NETROM:
		sp.nr->nr_family = AF_NETROM;
		memcpy(sp.nr->nr_addr.user_call,&Mycall,AXALEN);
		memcpy(sp.nr->nr_addr.node_call,&Mycall,AXALEN);
		bind(s,sp.p,sizeof(struct sockaddr_nr));
		break;
#endif
	}
}

unsigned
xgetdtablesize()
{
    return USOCKBASE+DEFNSOCK;
}

int
select(nfds, rfds, wfds, efds, timeout)
int nfds;
fd_set *rfds,*wfds,*efds;
struct timeval *timeout;
{
    int i,secs,lim,nos;
    extern int *Selectcnt;
    struct usock *up;
    void (*oldhandler)();
    static void selhandler();
    struct sembuf sembuf;
    fd_set rf, wf, ef;
    time_t starttime, remaining;
    lim = (nfds > getdtablesize()) ? getdtablesize() : nfds;
    errno = EBADF;
    if (rfds != (fd_set *)0)
	 for (i=0; i < lim; ++i)
	     if (FD_ISSET(i,rfds))
		 return -1;
    if (wfds != (fd_set *)0)
	 for (i=0; i < lim; ++i)
	     if (FD_ISSET(i,wfds))
		 return -1;
    if (efds != (fd_set *)0)
	 for (i=0; i < lim; ++i)
	     if (FD_ISSET(i,efds))
		 return -1;
    if(nfds <= getdtablesize())
	return -1;

    if(timeout == (struct timeval *)0)
	secs = -1;
    else secs = timeout->tv_sec;
    FD_ZERO(&rf);
    FD_ZERO(&wf);
    FD_ZERO(&ef);
    nos = 0;
    for(;;) {
	if(rfds != (fd_set *)0)
	    for(i=getdtablesize(); i < nfds; ++i)
		if(FD_ISSET(i,rfds)) {
		    if((up = itop(i)) == NULLUSOCK)
			return -1;
		    if(socklen(i,0) > 0 || up->rdysock != -1) {
			FD_SET(i,&rf);
			++nos;
		    }
		}
	if(wfds != (fd_set *)0)
	    for(i=getdtablesize(); i < nfds; ++i) {
		if((up = itop(i)) == NULLUSOCK)
		    return -1;
		if(FD_ISSET(i,wfds) && txok(i)) {
		    FD_SET(i,&wf);
		    ++nos;
		}
	    }
	/* Any exceptional conditions are not yet implemented */
	if(efds != (fd_set *)0)
	    for(i=getdtablesize(); i < nfds; ++i)
		if((up = itop(i)) == NULLUSOCK)
		    return -1;
	if(nos || secs == 0) {
	    if(rfds != (fd_set *)0)
		*rfds = rf;
	    if(wfds != (fd_set *)0)
		*wfds = wf;
	    if(efds != (fd_set *)0)
		*efds = ef;
	    return nos;
	}
	holdusock();
	++*Selectcnt;
	releaseusock();
	Alarmexp = 0;
	if(secs != -1) {
	    starttime = time((time_t *)0);
	    oldhandler = signal(SIGALRM,selhandler);
	    alarm(secs);
	}
	sembuf.sem_num = SEM_SELECT;
	sembuf.sem_op = WAITDEC_SEM;
	sembuf.sem_flg = 0;
	if (semop(Mainmem.semid,&sembuf,1) == -1) { 
	    holdusock();
	    if(*Selectcnt)
		--*Selectcnt;
	    releaseusock();
	    if(rfds != (fd_set *)0)
		*rfds = rf;
	    if(wfds != (fd_set *)0)
		*wfds = wf;
	    if(efds != (fd_set *)0)
		*efds = ef;
	    if(Alarmexp)
		return 0;		 /* the alarm timer expired */
	    else
		return -1;
	}
	if(secs != -1) {
	    alarm(0);
	    signal(SIGALRM,oldhandler);
	    remaining = time((time_t *)0) - starttime - secs;
	    if(remaining > 0)
		secs = remaining;
	    else
		secs = 0;
	}
    }
}

static void
selhandler()
{
	Alarmexp = 1;
}

static int
txok(s)
int s;
{
    struct usock *up;
    if((up = itop(s)) == NULLUSOCK)
	return 0;
    if(up->cb.p == NULLCHAR)
	return 0;
    switch(up->type){
    case TYPE_TCP:
	if(up->cb.tcb->sndcnt > up->cb.tcb->window
	   || up->cb.tcb->state != SYN_RECEIVED)
	    return 0;
	break;
    case TYPE_AX25I:
	if(len_q(up->cb.ax25->txq) > up->cb.ax25->window ||
	   (up->cb.ax25->state != CONNECTED && up->cb.ax25->state != RECOVERY))
	    return 0;
	break;
    case TYPE_NETROML4:
	if(up->cb.nr4->nbuffered >= up->cb.nr4->window
	   || up->cb.nr4->state != NR4STCON)
	    return 0;
	break;
    }
    return 1;
}

#ifndef SIOCSPGRP
#define SIOCSPGRP	900
#endif
#ifndef FIONREAD
#define FIONREAD	901
#endif
#ifndef FIONBIO
#define FIONBIO		902
#endif
#ifndef FASYNC
#define FASYNC		903
#endif
int
xioctl(s,request,arg)
int s,request;
void *arg;
{
    struct usock *up;
    int intarg;
    if(s < getdtablesize())
	return ioctl(s,request,arg);
    if((up = itop(s)) == NULLUSOCK){
	errno = EBADF;
	return -1;
    }
    switch(request) {
    case I_SETSIG:
	intarg = (int) arg;
	if(intarg == 0 && up->sigpid == 0) {
	    errno = EINVAL;
	    return -1;
	}
	if((up->sigpid > 0 && up->sigpid != getpgrp(getpid()))
	   || (up->sigpid < 0 && up->sigpid != -getpid())){
	    errno = EAGAIN;
	    return -1;
	}
	up->sigarg = intarg;
	if(intarg) {
	    if(up->sigpid == 0)
		up->sigpid = -getpid();
	}
	else
	    up->sigpid = 0;
	return 0;
    case I_GETSIG:
	if(arg == NULL) {
	    errno = EFAULT;
	    return -1;
	}
	if((up->sigpid > 0 && up->sigpid != getpgrp(getpid()))
	   || (up->sigpid < 0 && up->sigpid != -getpid())){
	    errno = EAGAIN;
	    return -1;
	}
	*(int *)arg = (int) up->sigarg;
	return 0;
    case SIOCSPGRP:
	if(arg == NULL) {
	    errno = EFAULT;
	    return -1;
	}
	if((up->sigpid > 0 && up->sigpid != getpgrp(getpid()))
	   || (up->sigpid < 0 && up->sigpid != -getpid())){
	    errno = EAGAIN;
	    return -1;
	}
	up->sigpid = *(int *)arg;
	return 0;
    case FIONREAD:
	if(arg == NULL) {
	    errno = EFAULT;
	    return -1;
	}
	*(long *)arg = (long) socklen(s,0);
	return 0;
    case FIONBIO:
	if(arg == NULL) {
	    errno = EFAULT;
	    return -1;
	}
	if(*(int *)arg == 1)
	    up->noblock = 1;
	else
	    if(*(int *)arg == 0)
		up->noblock = 0;
	    else {
		errno = EINVAL;
		return -1;
	    }
	return 0;
    }
    errno = EINVAL;
    return -1;
}

int
xfcntl(s,cmd,arg)
int s,cmd,arg;
{
    int val = -1;
    if(s < getdtablesize())
	return fcntl(s,cmd,arg);
    switch(cmd){
    case F_GETFL:
	xioctl(s,I_GETSIG,&val);
	return val;
    case F_SETFL:
	if(xioctl(s,I_GETSIG,&val) != -1){
	    if(arg & FASYNC)
		val |= S_INPUT;
	    else
		val &= ~S_INPUT;
	    return xioctl(s,I_SETSIG,val);
	}
	return -1;
    }
    errno = EINVAL;
    return -1;
}

int
xsetsockopt(s,level,optname,optval,optlen)
int s,level,optname;
char *optval;    
int optlen;
{
#ifndef SYS5
    if(s < getdtablesize())
	return setsockopt(s,level,optname,optval,optlen);
#endif
    if(itop(s))
	    return 0;
	else return -1;
}
#ifdef SYS5
struct servent *
getservbyname(service,proto)
char *service, *proto;
{
    static struct servent sp;
    sp.s_aliases = (char **) 0;
    sp.s_name = (char *) 0;
    if(!strcmp(proto,"tcp")) {
	sp.s_proto = "tcp";
	if(!strcmp(service,"telnet")) {
	    sp.s_name = "telnet";
	    sp.s_port = IPPORT_TELNET;
	}
	if(!strcmp(service,"finger")) {
	    sp.s_name = "finger";
	    sp.s_port = IPPORT_FINGER;
	}
	if(!strcmp(service,"ftp")) {
	    sp.s_name = "ftp";
	    sp.s_port = IPPORT_FTP;
	}
	if(!strcmp(service,"ftp-data")) {
	    sp.s_name = "ftp-data";
	    sp.s_port = IPPORT_FTPD;
	}
	if(!strcmp(service,"smtp")) {
	    sp.s_name = "smtp";
	    sp.s_port = IPPORT_SMTP;
	}
	if(!strcmp(service,"echo")) {
	    sp.s_name = "echo";
	    sp.s_port = IPPORT_ECHO;
	}
	if(!strcmp(service,"discard")) {
	    sp.s_name = "discard";
	    sp.s_port = IPPORT_DISCARD;
	}
    }
    if(!strcmp(proto,"udp")) {
	sp.s_proto = "udp";
	if(!strcmp(service,"domain")) {
	    sp.s_name = "domain";
	    sp.s_port = IPPORT_DOMAIN;
	}
    }
    if(sp.s_name)
	return &sp;
    else
	return (struct servent *)0;
}
#endif
int
xfork()
{
    extern int Intoff;
    int pid,parent;
    int i,j,k;
    struct usock *up;
    if (!Initok && shminit() == -1)
	return fork();
    parent = getpid();
    holdusock();
    pid = fork();
    if (pid == -1)
	return -1;
    if (pid == 0) {
    	Intoff = 0;
	for(i=0; i < Nusock; ++i) {
	    up = &Usock[i];
	    if (up->type != NOTUSED)
		for(j=0; j < sizeof(up->pid)/sizeof(int); ++j)
		    if (up->pid[j] == parent) {
			for(k=0; k < sizeof(up->pid)/sizeof(int); ++k)
			    if (up->pid[k] == 0) {
				up->pid[k] = getpid();
				break;
			    }
			if (k == sizeof(up->pid)/sizeof(int))
			    fprintf(stderr,"xfork: too many processes using same file\n");
			break;
		    }
	}
	releaseusock();			/* signal that we are all done */
	return 0;
    }
    else
    {
	holdusock();			/* wait until the child is done */
	releaseusock();
	return pid;
    }
}
int
xvfork()
{
    extern int Intoff;
    int pid,parent;
    int i,j,k;
    struct usock *up;
    if (!Initok && shminit() == -1)
	return vfork();
    parent = getpid();
    holdusock();
    pid = vfork();
    if (pid == -1)
	return -1;
    if (pid == 0) {
    	Intoff = 0;
	for(i=0; i < Nusock; ++i) {
	    up = &Usock[i];
	    if (up->type != NOTUSED)
		for(j=0; j < sizeof(up->pid)/sizeof(int); ++j)
		    if (up->pid[j] == parent) {
			for(k=0; k < sizeof(up->pid)/sizeof(int); ++k)
			    if (up->pid[k] == 0) {
				up->pid[k] = getpid();
				break;
			    }
			if (k == sizeof(up->pid)/sizeof(int))
			    fprintf(stderr,"xvfork: too many processes using same file\n");
			break;
		    }
	}
	releaseusock();			/* signal that we are all done */
	return 0;
    }
    else
    {
	holdusock();			/* wait until the child is done */
	releaseusock();
	return pid;
    }
}

int
client_sendtcp(tcb, bp)
struct tcb *tcb;
struct mbuf *bp;
{
    int cnt;
    if(tcb == NULLTCB || bp == NULLBUF) {
	free_p(bp);
	return -1;
    }
    cnt = len_mbuf(bp);
    rpcall(RPC_SENDTCP,tcb,bp);
    return cnt;
}
int
client_recvtcp(tcb,bpp,len)
struct tcb *tcb;
struct mbuf **bpp;
int len;
{
    if(tcb == NULLTCB || bpp == NULLBUFP)
	return -1;
    if(tcb->rcvcnt == 0)
	switch(tcb->state){
	case CLOSED:
	case CLOSE_WAIT:
	case CLOSING:
	case LAST_ACK:
	case TIME_WAIT:
	    *bpp = NULLBUF;
	    return 0;
	}
    rpcall(RPC_RECVTCP,tcb,bpp,len);
    if(*bpp == NULLBUF)
	return -1;
    else
	return len_mbuf(*bpp);
}

