#include <stdio.h>
#include <termio.h>
#include <signal.h>
#include "global.h"
#include <ctype.h>
#include "mbuf.h"
#include "socket.h"
#include "telnet.h"
#include "tty.h"

#define CMD_MODE        1       /* Command mode */
#define CONV_MODE       2       /* Converse mode */
static int Mode,Savedmode,Whoiswho;
static char Escape = 0x1d;	/* default escape character is ^] */
static char *Prog;
static struct telnet *tnp;
static struct termio Savecon;
static void doopt(),dontopt(),willopt(),wontopt(),answer(),reset_io();

int Refuse_echo = 0;
int Unix_line_mode = 0;    /* if true turn <cr> to <nl> when in line mode */

#ifdef	DEBUG
char *T_options[] = {
	"Transmit Binary",
	"Echo",
	"",
	"Suppress Go Ahead",
	"",
	"Status",
	"Timing Mark"
};
#endif

main(argc,argv)
int argc;
char *argv[];
{
    int s;
    char *cp;
    Prog = argv[0];
    if (argc < 2) {
	fprintf(stderr,"Usage: %s <address> [<port>]\n",Prog);
	exit(1);
    }
    if ((s = callsetup(argc,argv)) == -1) {
	iostop();
	exit(1);
    }
    set_io();
    /* Fork off the receiver process */
    if ((Whoiswho = xfork()) == 0) {
	sockreader(s);
	exit(0);
    }
    keyboard(s);
    cp = sockerr(s);
    printf("Connection closed: %s\n",cp != NULLCHAR ? cp : "EOF");
    close_s(s);
    reset_io();
}

/* Execute user telnet command */
int
callsetup(argc,argv)
int argc;
char *argv[];
{
	struct sockaddr_in fsocket;
	char *cp;
	int s;

	fsocket.sin_family = AF_INET;
	if(argc < 3)
		fsocket.sin_port = IPPORT_TELNET;
	else
		fsocket.sin_port = atoi(argv[2]);

	if((fsocket.sin_addr.s_addr = resolve(argv[1])) == 0){
		fprintf(stderr,"Bad hostname, %s\n",argv[1]);
		return -1;
	}
	if((s = socket(AF_INET,SOCK_STREAM,0)) == -1){
		fprintf(stderr,"Can't create socket: errno %d\n",errno);
		return -1;
	}
	printf("Trying %s ...\n",psocket((struct sockaddr *)&fsocket));
	if(connect(s,(char *)&fsocket,SOCKSIZE) == -1){
		cp = sockerr(s);
		printf("Telnet connect failed: %s\n",
		 cp != NULLCHAR ? cp : "");
		return -1;
	}
	printf("Connected to %s.\nEscape character is '^%c'.\n",
		argv[1],'@'+Escape);

	/* Initialize shared Telnet protocol descriptor */
	tnp = (struct telnet *) shcalloc(1,sizeof(struct telnet));
	tnp->s = s;

	return s;
}

keyboard(s) 	/* Send whatever's typed on the terminal */
int s;
{
    struct mbuf *bp;
    int c;
    char *cp;

    Mode = CONV_MODE;
    ttysetmode(TTY_ECHO|TTY_EDIT);
    for(;;) {
	if ((c = getchar()) == EOF)
	    return;
	if (c == Escape && Mode == CONV_MODE) {
	    Savedmode = ttygetmode();
	    ttysetmode(TTY_ECHO|TTY_EDIT);
	    Mode = CMD_MODE;
	    printf("%s> ",Prog);
	    continue;
	}
	if ((bp = ttydriv(c)) != NULLBUF)
	    switch(Mode){
	    case CMD_MODE:
		if (keybcommand(bp,s) != 0)
		    return;
		if (Mode == CMD_MODE)
		    printf("%s> ",Prog);
		else
		    ttysetmode(Savedmode);
		break;
	    case CONV_MODE:
		/* If we're doing our own echoing and recording is enabled,
		 * record it
		 */
		if(!tnp->remote[TN_ECHO] && tnp->Record != NULLFILE)
			write_p(tnp->Record,bp);

		if (send_mbuf(s,bp,0,NULLCHAR,0) == -1)
		    return;
		break;
	    }
    }
}

/* User telnet receive task, started by user telnet command */
sockreader(s)
int s;
{
	int c;
	char *cp;
	struct mbuf *bp = NULLBUF;

	while((c = recvchar(s,&bp)) != -1){

		if(uchar(c) == IAC){
			/* IAC received, get command sequence */
			c = recvchar(s,&bp);
			switch(uchar(c)){
			case WILL:
				willopt(tnp,recvchar(s,&bp));
				break;
			case WONT:
				wontopt(tnp,recvchar(s,&bp));
				break;
			case DO:
				doopt(tnp,recvchar(s,&bp));
				break;
			case DONT:
				dontopt(tnp,recvchar(s,&bp));
				break;
			case IAC:	/* Escaped IAC */
				putchar(IAC);
				if(tnp->Record != NULLFILE)
					putc(IAC,tnp->Record);
				break;
			}
		} else {
			/* Ordinary character */
			if(!tnp->remote[TN_TRANSMIT_BINARY])
				c &= 0x7f;

			putchar(c);
			if(tnp->Record != NULLFILE)
				putc(c,tnp->Record);
		}
	}
	/* Close seen from remote host */
	free_p(bp);
}

/* parse commands, quick and dirty */
int
keybcommand(bp,s)
struct mbuf *bp;
int s;
{
    char buf[80], *p;

    dqdata(bp,buf,80);
    p = buf;
    while (*p == ' ' || *p == '\t')
	++p;
    if (isupper (*p))
	*p = tolower(*p);
    if (*p == '\n' || *p == '\r') {
	Mode = CONV_MODE;
	return 0;
    }
    if (*p == '?' || *p == 'h') {
	printf("Commands may be abbreviated.  Commands are:\n\n");
	printf("kick\t\tkick connection\n");
	printf("quit\t\texit program\n");
	printf("record <file>\trecord into <file>\n");
	printf("z\t\tsuspend program\n");
	return 0;
    }
    if (*p == 'k') {
	sockkick(s);
	Mode = CONV_MODE;
	return 0;
    }
    if (*p == 'z') {
#ifdef SYS5
	printf("Not yet implemented\n");
#else
	kill(getpid(),SIGSTOP);
	set_io();
	Mode = CONV_MODE;
#endif
	return 0;
    }
    if (*p == 'q') {
	if (tnp->Record)
	    fclose(tnp->Record);
	return -1;
    }
    if (*p == 'r') {
	p = strchr(p,' ');
	if (p) {
	    if (!strcmp(p,"off")){
		fclose(tnp->Record);
		tnp->Record = NULLFILE;
	    }
	    else
		if ((tnp->Record = fopen(p,"a")) == NULLFILE)
		    printf("Could not open %s for writing\n",p);
	}
	Mode = CONV_MODE;
	return 0;
    }
    printf("?Invalid command.\n");
    return 0;
}
/* The guts of the actual Telnet protocol: negotiating options */
static
void
willopt(tn,opt)
struct telnet *tn;
char opt;
{
	int ack;

#ifdef	DEBUG
	printf("recv: will ");
	if(uchar(opt) <= NOPTIONS)
		printf("%s\n",T_options[opt]);
	else
		printf("%u\n",opt);
#endif
	
	switch(uchar(opt)){
	case TN_TRANSMIT_BINARY:
	case TN_ECHO:
	case TN_SUPPRESS_GA:
		if(tn->remote[uchar(opt)] == 1)
			return;		/* Already set, ignore to prevent loop */
		if(uchar(opt) == TN_ECHO){
			if(Refuse_echo){
				/* User doesn't want to accept */
				ack = DONT;
				break;
			} else
				ttysetmode(0);		/* Put tty into raw mode */
		}
		tn->remote[uchar(opt)] = 1;
		ack = DO;			
		break;
	default:
		ack = DONT;	/* We don't know what he's offering; refuse */
	}
	answer(tn,ack,opt);
}
static
void
wontopt(tn,opt)
struct telnet *tn;
char opt;
{
#ifdef	DEBUG
	printf("recv: wont ");
	if(uchar(opt) <= NOPTIONS)
		printf("%s\n",T_options[uchar(opt)]);
	else
		printf("%u\n",uchar(opt));
#endif
	if(uchar(opt) <= NOPTIONS){
		if(tn->remote[uchar(opt)] == 0)
			return;		/* Already clear, ignore to prevent loop */
		tn->remote[uchar(opt)] = 0;
		if(uchar(opt) == TN_ECHO)
			ttysetmode(TTY_ECHO|TTY_EDIT);	/* Put tty into cooked mode */
	}
	answer(tn,DONT,opt);	/* Must always accept */
}
static
void
doopt(tn,opt)
struct telnet *tn;
char opt;
{
	int ack;

#ifdef	DEBUG
	printf("recv: do ");
	if(uchar(opt) <= NOPTIONS)
		printf("%s\n",T_options[uchar(opt)]);
	else
		printf("%u\n",uchar(opt));
#endif
	switch(uchar(opt)){
#ifdef	FUTURE	/* Use when local options are implemented */
		if(tn->local[uchar(opt)] == 1)
			return;		/* Already set, ignore to prevent loop */
		tn->local[uchar(opt)] = 1;
		ack = WILL;
		break;
#endif
	default:
		ack = WONT;	/* Don't know what it is */
	}
	answer(tn,ack,opt);
}
static
void
dontopt(tn,opt)
struct telnet *tn;
char opt;
{
#ifdef	DEBUG
	printf("recv: dont ");
	if(uchar(opt) <= NOPTIONS)
		printf("%s\n",T_options[uchar(opt)]);
	else
		printf("%u\n",uchar(opt));
#endif
	if(uchar(opt) <= NOPTIONS){
		if(tn->local[uchar(opt)] == 0){
			/* Already clear, ignore to prevent loop */
			return;
		}
		tn->local[uchar(opt)] = 0;
	}
	answer(tn,WONT,opt);
}
static
void
answer(tn,r1,r2)
struct telnet *tn;
int r1,r2;
{
	char s[3];

#ifdef	DEBUG
	switch(r1){
	case WILL:
		printf("sent: will ");
		break;
	case WONT:
		printf("sent: wont ");
		break;
	case DO:
		printf("sent: do ");
		break;
	case DONT:
		printf("sent: dont ");
		break;
	}
	if(r2 <= 6)
		printf("%s\n",T_options[r2]);
	else
		printf("%u\n",r2);
#endif

	s[0] = IAC;
	s[1] = r1;
	s[2] = r2;
	send(tn->s,s,3,0);
}
set_io()
{
    struct termio ttybuf;

    /* We want unbuffered output */
    freopen(ttyname(1),"w",stdout);
    setbuf(stdout,NULLCHAR);
    ioctl(fileno(stdin), TCGETA, &ttybuf);
    Savecon = ttybuf;
    ttybuf.c_iflag &= ~IXON;
    ttybuf.c_lflag &= ~(ICANON|ECHO);
    ttybuf.c_cc[VTIME] = 0;
    ttybuf.c_cc[VMIN] = 1;
    ioctl(fileno(stdin), TCSETAW, &ttybuf);
    signal(SIGINT, reset_io);
    signal(SIGTERM, reset_io);
    signal(SIGHUP, reset_io);
    signal(SIGQUIT, reset_io);
}
static void
reset_io()
{
    if (Whoiswho != 0) {
	shfree((char *)tnp);
	kill(Whoiswho,SIGTERM);
    }
    ioctl(fileno(stdin), TCSETAW, &Savecon);
    iostop();
}
