#ifndef	TIMER_STOP

#include "global.h"

/* Software timers
 * There is one of these structures for each simulated timer.
 * Whenever the timer is running, it is on a doubly-linked list
 * pointed to by "Timers". The list is sorted in ascending order of
 * expiration, with the next timer to expire at the head. Each timer's
 * count field contains only the additional INCREMENT over all preceeding
 * timers; this allows the hardware timer interrupt to decrement only the
 * first timer on the chain until it hits zero.
 *
 * Stopping a timer or letting it expire causes it to be removed
 * from the list. Starting a timer puts it on the list at the right
 * place. These operations have to be careful about adjusting the count
 * field of the next entry in the list.
 */
struct timer {
	struct timer *next;	/* Doubly-linked-list pointers */
	struct timer *prev;
	int32 start;		/* Period of counter (load value) */
	int32 count;		/* Ticks to go until expiration */
	void (*func)();		/* Function to call at expiration */
	char *arg;		/* Arg to pass function */
	char state;		/* Timer state */
#define	TIMER_STOP	0
#define	TIMER_RUN	1
#define	TIMER_EXPIRE	2
};
#define	NULLTIMER	(struct timer *)0
#define	MAX_TIME	(int32)4294967295	/* Max long integer */
#define	MSPTICK		200		/* Milliseconds per tick */
/* Useful user macros that hide the timer structure internals */
#define	set_timer(t,x)	(((t)->start) = (x)/MSPTICK)
#define	dur_timer(t)	((t)->start)
#define	run_timer(t)	((t)->state == TIMER_RUN)

extern int32 Clock;	/* Count of ticks since start up */

#if	defined(__STDC__) || defined(__TURBOC__)
void alarm(int32 ticks);
void start_timer(struct timer *t);
void stop_timer(struct timer *t);
int32 read_timer(struct timer *t);
void pause(int32 ticks);
#else
void alarm();
void start_timer(),stop_timer();
int32 read_timer();
void pause();
#endif

#endif	/* TIMER_STOP */


