#ifndef	NUDP

#include "global.h"
#include "netuser.h"

/* User Datagram Protocol definitions */

#define	NUDP	20

/* Structure of a UDP protocol header */
struct udp {
	int16 source;	/* Source port */
	int16 dest;	/* Destination port */
	int16 length;	/* Length of header and data */
	int16 checksum;	/* Checksum over pseudo-header, header and data */
};
#define	UDPHDR	8	/* Length of UDP header */

/* User Datagram Protocol control block
 * Each entry on the receive queue consists of the
 * remote socket structure, followed by any data
 */
struct udp_cb {
	struct udp_cb *prev;	/* Linked list pointers */
	struct udp_cb *next;
	struct socket socket;	/* Local port accepting datagrams */
	void (*r_upcall)();	/* Function to call when one arrives */
	struct mbuf *rcvq;	/* Queue of pending datagrams */
	int rcvcnt;		/* Count of pending datagrams */
	int user;		/* User link */
};
extern struct udp_cb *Udps[];	/* Hash table for UDP structures */
#define	NULLUDP	(struct udp_cb *)0

/* UDP statistics counters */
struct udp_stat {
	int16 rcvd;		/* Packets received */
	int16 sent;		/* Packets sent */
	int16 cksum;		/* Checksum errors */
	int16 unknown;		/* Unknown socket */
	int16 bdcsts;		/* Incoming broadcasts */
};
extern struct udp_stat Udp_stat;

/* UDP primitives */
#if	defined(__STDC__) || defined(__TURBOC__)
void udp_input(struct mbuf *bp,struct ip *ip,int rxbroadcast);
int recv_udp(struct udp_cb *up,struct socket *fsocket,struct mbuf **bp);
int send_udp();
int del_udp(struct udp_cb *up);
struct udp_cb *open_udp();
struct mbuf *htonudp(struct udp *udp,struct mbuf *data,struct pseudo_header *ph);
int ntohudp(struct udp *udp,struct mbuf **bpp);
#else
void udp_input();
int recv_udp(),send_udp(),del_udp();
void udp_dump();
struct udp_cb *open_udp();
struct mbuf *htonudp();
int ntohudp();
#endif

#endif	/* NUDP */

