/* OS- and machine-dependent stuff for 4.[23] BSD UNIX */

/*
	FILE: UNIX.io.c
	
	Routines:
		asy_init()
		asy_stop()
		asy_speed()
		asy_output()
		doasystat()
		asy_ioctl()
	Written or converted by Mikel Matthews, N9DVG, and others
	
*/

#include <stdio.h>
#include <stropts.h>
#include <sys/types.h>
#include <sys/file.h>
#include <termio.h>
#include <fcntl.h>
#include <signal.h>
#include "global.h"
#include "mbuf.h"
#include "iface.h"
#include "unix.h"
#include "cmdparse.h"

#ifndef	B19200
#define	B19200	EXTA
#endif

#ifndef	B38400
#define	B38400	EXTB
#endif

struct asy Asy[ASY_MAX];
struct iface *Ifaces;

/* Initialize asynch port "dev" */
int
asy_init(dev,iface,arg1,arg2,bufsize)
int16 dev;
struct iface *iface;
char *arg1,*arg2;
unsigned bufsize;
{
	register struct asy *ap;
	extern struct iface *Ifaces;
	struct termio sgttyb,mysavetty;
	
	if (dev >= Nasy)
		return(-1);

	ap = &Asy[dev];
	ap->tty = malloc(strlen(arg2)+1);
	strcpy(ap->tty, arg2);
	ap->iface = iface;
	printf("asy_init: tty name = %s\n", ap->tty);

#ifdef SYS5
	sighold(SIGALRM); /* to prevent interrupted system calls */
#endif
	if ((ap->IORser = open (ap->tty, O_RDWR, 0)) < 0)
	{
		perror ("Could not open device");
		return (-1);
	}
#ifdef SYS5
	sigrelse(SIGALRM);
#endif

 /* 
  * get the stty structure and save it 
  */

	if (ioctl (ap->IORser, TCGETA, &mysavetty) < 0)
	{
		perror ("ioctl failed on device");
		return (-1);
	}
 /* 
  * copy over the structure 
  */

	sgttyb = mysavetty;
	sgttyb.c_iflag = (IGNBRK | IGNPAR);
	sgttyb.c_oflag = 0;
	sgttyb.c_lflag = 0;
	sgttyb.c_cflag = (B9600 | CS8 | CREAD);
	sgttyb.c_cc[VTIME] = 0;
	sgttyb.c_cc[VMIN] = 1;
 /* 
  * now set the modes and flags 
  */
	if (ioctl (ap->IORser, TCSETAF, &sgttyb) < 0)
	{
		perror ("ioctl could not set parameters for IORser");
		return (-1);
	}
	return (0);
}

int
asy_stop(dev)
int dev;
{
}

/* Set asynch line speed */
int
asy_speed(dev,speed)
int16 dev;
int speed;
{
	struct termio sgttyb;

	if(speed == 0 || dev >= Nasy)
		return(-1);

	if (ioctl (Asy[dev].IORser, TCGETA, &sgttyb) < 0)
	{
		perror ("ioctl could not get parameters");
		return (-1);
	}

	sgttyb.c_cflag &= ~CBAUD;

	switch ((unsigned)speed) {
	case 0:
		sgttyb.c_cflag |= B0;
		break;
	case 50:
		sgttyb.c_cflag |= B50;
		break;
	case 75:
		sgttyb.c_cflag |= B75;
		break;
	case 110:
		sgttyb.c_cflag |= B110;
		break;
	case 134:
		sgttyb.c_cflag |= B134;
		break;
	case 150:
		sgttyb.c_cflag |= B150;
		break;
	case 200:
		sgttyb.c_cflag |= B200;
		break;
	case 300:
		sgttyb.c_cflag |= B300;
		break;
	case 600:
		sgttyb.c_cflag |= B600;
		break;
	case 1200:
		sgttyb.c_cflag |= B1200;
		break;
	case 1800:
		sgttyb.c_cflag |= B1800;
		break;
	case 2400:
		sgttyb.c_cflag |= B2400;
		break;
	case 4800:
		sgttyb.c_cflag |= B4800;
		break;
	case 9600:
		sgttyb.c_cflag |= B9600;
		break;
	case 19200:
		sgttyb.c_cflag |= B19200;
		break;
	case 38400:
		sgttyb.c_cflag |= B38400;
		break;
	default:
		printf("asy_speed: Unknown speed (%d)\n", speed);
		break;
	}

	if (ioctl (Asy[dev].IORser, TCSETAW, &sgttyb) < 0)
	{
		perror ("ioctl could not set parameters for IORser");
		return (-1);
	}
	Asy[dev].speed = speed;
	return(0);
}
/* Send a buffer to serial transmitter */
asy_output(dev,buf,cnt)
unsigned dev;
char *buf;
unsigned short cnt;
{
	int i_state;

	if(dev >= Nasy)
		return(-1);
	i_state = dirps();
	if ( write(Asy[dev].IORser, buf, (int)cnt) < cnt)
	{
		perror("asy_output");
		printf("asy_output: error in writing to device %d\n", dev);
		restore (i_state);
		return(-1);
	}
	restore (i_state);
	Asy[dev].txchar += cnt;
	return(0);
}
asy_ioctl(Iface, argc, argv)
struct iface *Iface;
int	argc;
char	*argv[];
{
	if (argc < 1) {
		printf("%d\r\n", Asy[Iface->dev].speed);
		return 0;
	}
	return asy_speed(Iface->dev, atoi(argv[0]));
}

int
doasystat(argc,argv)
int argc;
char *argv[];
{
	register struct asy *asyp;

	for(asyp = Asy;asyp < &Asy[Nasy];asyp++)
	    if (asyp->tty != NULLCHAR)
		printf("%s (%s %d baud): rxchar %lu txchar %lu\n",
		 asyp->iface->name,asyp->tty,asyp->speed,asyp->rxchar,asyp->txchar);
	return 0;
}
