#include <stropts.h>
#include <signal.h>
#include "global.h"
#include "mbuf.h"
#include "memory.h"
#include "upcalls.h"
#include "socket.h"
#include "usock.h"
#include "config.h"
#include "socket.h"
#include "iface.h"
#include "ip.h"
#include "tcp.h"
#include "udp.h"
#include "lapb.h"
#include "nr4.h"
    
static struct usock *itop();
struct usock *Usock;
int Nusock = DEFNSOCK;		/* Number of socket entries */
void s_arcall(),s_atcall(),s_ascall(),rip_recv();
static void s_nrcall(),s_ntcall(),s_nscall();
static void s_trcall(),s_ttcall(),s_tscall(),s_urcall(),trdiscard();

void (*uptab[])() = {
    s_arcall,
    s_atcall,
    s_ascall,
    s_nrcall,
    s_ntcall,
    s_nscall,
    s_trcall,
    s_ttcall,
    s_tscall,
    s_urcall,
    trdiscard,
    rip_recv
};

/* The following two variables are needed because there can be only one
 * socket listening on each of the AX.25 modes (I and UI)
 */
#ifdef	AX25
int Axi_sock = -1;	/* Socket number listening for AX25 connections */
int Axui_sock = -1;	/* Socket number listening for AX25 UI frames */
struct mbuf *Bcq;	/* Queue of incoming UI frames */
extern int16 Axwindow;	/* Window size in bytes */

/* Function that handles incoming UI frames from lapb.c */
void
beac_input(iface,src,bp)
struct iface *iface;
struct ax25_addr *src;
struct mbuf *bp;
{
	struct mbuf *hdr;
	struct sockaddr_ax *sax;

	if(Axui_sock == -1){
		/* Nobody there to read it */
		free_p(bp);
	} else {
		if((hdr = pushdown(NULLBUF,sizeof(struct sockaddr_ax))) == NULLBUF){
			free_p(bp);
			return;
		}
		sax = (struct sockaddr_ax *)hdr->data;
		sax->sax_family = AF_AX25;
		ASSIGN(sax->ax25_addr,*src);
		strncpy((char *)sax->iface,iface->name,ILEN);
		hdr->next = bp;
		enqueue(&Bcq,hdr);
		psignal(&Bcq,1);
	}
}
#endif

/* Raw IP receive upcall routine */
void
rip_recv(rp)
struct raw_ip *rp;
{
	struct usock *up;
	up = itop(rp->user);
	psignal(up,1);
	if(up != NULLUSOCK && (up->sigarg & S_INPUT))
	    if(up->sigpid > 0)
		killpg(up->sigpid,SIGPOLL);
	    else
		kill(-up->sigpid,SIGPOLL);
	pwait(NULL);
}

/* UDP receive upcall routine */
static void
s_urcall(udp,cnt)
struct udp_cb *udp;
int cnt;
{
	struct usock *up;
	up = itop(udp->user);
	psignal(up,1);
	if(up != NULLUSOCK && (up->sigarg & S_INPUT))
	    if(up->sigpid > 0)
		killpg(up->sigpid,SIGPOLL);
	    else
		kill(-up->sigpid,SIGPOLL);
	pwait(NULL);
}
/* TCP receive upcall routine */
static void
s_trcall(tcb,cnt)
struct tcb *tcb;
int cnt;
{
	/* Wake up anybody waiting for data, and let them run */
	struct usock *up;
	up = itop(tcb->user);
	psignal(up,1);
	if(up != NULLUSOCK && (up->sigarg & S_INPUT))
	    if(up->sigpid > 0)
		killpg(up->sigpid,SIGPOLL);
	    else
		kill(-up->sigpid,SIGPOLL);
	pwait(NULL);
}
/* TCP transmit upcall routine */
static void
s_ttcall(tcb,cnt)
struct tcb *tcb;
int cnt;
{
	struct usock *up;
	up = itop(tcb->user);
	/* Wake up anybody waiting to send data, and let them run */
	psignal(up,1);
	if(up != NULLUSOCK && (up->sigarg & S_OUTPUT))
	    if(up->sigpid > 0)
		killpg(up->sigpid,SIGPOLL);
	    else
		kill(-up->sigpid,SIGPOLL);
	pwait(NULL);
}
/* TCP state change upcall routine */
static void
s_tscall(tcb,old,new)
struct tcb *tcb;
int old,new;
{
	int s,ns;
	struct usock *up,*nup,*oup;
	union sp sp;
	char i_state;

	s = tcb->user;
	oup = up = itop(s);

printf("tscall, tcb %d new %d old %d\n",tcb,new,old);
	switch(new){
	case CLOSED:
		/* Clean up. If the user has already closed the socket,
		 * then up will be null (s was set to -1 by the close routine).
		 * If not, then this is an abnormal close (e.g., a reset)
		 * and clearing out the pointer in the socket structure will
		 * prevent any further operations on what will be a freed
		 * control block. Also wake up anybody waiting on events
		 * related to this tcb so they will notice it disappearing.
		 */
		i_state = dirps();
		if(up != NULLUSOCK){
			up->cb.tcb = NULLTCB;
			up->errcodes[0] = tcb->reason;
			up->errcodes[1] = tcb->type;
			up->errcodes[2] = tcb->code;
		}
		restore(i_state);
		del_tcp(tcb);
		break;
	case SYN_RECEIVED:
		/* Handle an incoming connection. If this is a server TCB,
		 * then we're being handed a "clone" TCB and we need to
		 * create a new socket structure for it. In either case,
		 * find out who we're talking to and wake up the guy waiting
		 * for the connection.
		 */
		i_state = dirps();
		if(tcb->flags.clone){
			/* Clone the socket */
			ns = socket(AF_INET,SOCK_STREAM,0);
			nup = itop(ns);
			ASSIGN(*nup,*up);
			tcb->user = ns;
			nup->cb.tcb = tcb;
			/* Allocate new memory for the name areas */
			nup->name = shmalloc(SOCKSIZE);
			nup->peername = shmalloc(SOCKSIZE);
			/* Store the new socket # in the old one */
			up->rdysock = ns;
			up = nup;
			s = ns;
		} else {
			/* Allocate space for the peer's name */
			up->peername = shmalloc(SOCKSIZE);
			/* Store the old socket # in the old socket */
			up->rdysock = s;
		}
		/* Load the addresses. Memory for the name has already
		 * been allocated, either above or in the original bind.
		 */
		sp.p = up->name;
		sp.in->sin_family = AF_INET;
		sp.in->sin_addr.s_addr = up->cb.tcb->conn.local.address;
		sp.in->sin_port = up->cb.tcb->conn.local.port;
		up->namelen = SOCKSIZE;

		sp.p = up->peername;
		sp.in->sin_family = AF_INET;
		sp.in->sin_addr.s_addr = up->cb.tcb->conn.remote.address;
		sp.in->sin_port = up->cb.tcb->conn.remote.port;
		up->peernamelen = SOCKSIZE;

		restore(i_state);
		/* Wake up the guy accepting it, and let him run */
		psignal(oup,1);
		pwait(NULL);
		break;
	default:	/* Ignore all other state transitions */
		break;
	}
	psignal(up,0);	/* In case anybody's waiting */
}
/* Discard data received on a TCP connection. Used after a receive shutdown or
 * close_s until the TCB disappears.
 */
static void
trdiscard(tcb,cnt)
struct tcb *tcb;
int cnt;
{
	struct mbuf *bp;

	recv_tcp(tcb,&bp,cnt);
	free_p(bp);
}
#ifdef	AX25
/* AX.25 receive upcall */
void
s_arcall(axp,cnt)
struct ax25_cb *axp;
int cnt;
{
	int ns;
	struct usock *up,*nup,*oup;
	union sp sp;

	up = itop(axp->user);
	/* When AX.25 data arrives for the first time the AX.25 listener
	   is notified, if active. If the AX.25 listener is a server its
	   socket is duplicated in the same manner as in s_tscall().
	 */
	if (Axi_sock != -1 && axp->user == -1) {
		oup = up = itop(Axi_sock);
		/* From now on, use the same upcalls as the listener */
		axp->t_upcall = up->cb.ax25->t_upcall;
		axp->r_upcall = up->cb.ax25->r_upcall;
		axp->s_upcall = up->cb.ax25->s_upcall;
		if (up->cb.ax25->flags.clone) {
			/* Clone the socket */
			ns = socket(AF_AX25,SOCK_STREAM,0);
			nup = itop(ns);
			ASSIGN(*nup,*up);
			axp->user = ns;
			nup->cb.ax25 = axp;
			/* Allocate new memory for the name areas */
			nup->name = shmalloc(sizeof(struct sockaddr_ax));
			nup->peername = shmalloc(sizeof(struct sockaddr_ax));
			/* Store the new socket # in the old one */
			up->rdysock = ns;
			up = nup;
		} else {
			axp->user = Axi_sock;
			del_ax25(up->cb.ax25);
			up->cb.ax25 = axp;
			/* Allocate space for the peer's name */
			up->peername = shmalloc(sizeof(struct sockaddr_ax));
			/* Store the old socket # in the old socket */
			up->rdysock = Axi_sock;
		}
		/* Load the addresses. Memory for the name has already
		 * been allocated, either above or in the original bind.
		 */
		sp.p = up->name;
		sp.ax->sax_family = AF_AX25;
		ASSIGN(sp.ax->ax25_addr,axp->local);
		memcpy(sp.ax->iface,axp->iface->name,ILEN);
		up->namelen = sizeof(struct sockaddr_ax);

		sp.p = up->peername;
		sp.ax->sax_family = AF_AX25;
		ASSIGN(sp.ax->ax25_addr,axp->remote);
		memcpy(sp.ax->iface,axp->iface->name,ILEN);
		up->peernamelen = sizeof(struct sockaddr_ax);
		/* Wake up the guy accepting it, and let him run */
		up = oup;
	}
	/* Wake up anyone waiting, and let them run */
	psignal(up,1);
	if(up != NULLUSOCK && (up->sigarg & S_INPUT))
	    if(up->sigpid > 0)
		killpg(up->sigpid,SIGPOLL);
	    else
		kill(-up->sigpid,SIGPOLL);
	pwait(NULL);
}
/* AX.25 transmit upcall */
void
s_atcall(axp,cnt)
struct ax25_cb *axp;
int cnt;
{
	struct usock *up;
	up = itop(axp->user);
	/* Wake up anyone waiting, and let them run */
	psignal(up,1);
	if(up != NULLUSOCK && (up->sigarg & S_OUTPUT))
	    if(up->sigpid > 0)
		killpg(up->sigpid,SIGPOLL);
	    else
		kill(-up->sigpid,SIGPOLL);
	pwait(NULL);
}
/* AX25 state change upcall routine */
void
s_ascall(axp,old,new)
register struct ax25_cb *axp;
int old,new;
{
	int s;
	struct usock *up;

	s = axp->user;
	up = itop(s);

	switch(new){
	case DISCONNECTED:
		/* Clean up. If the user has already closed the socket,
		 * then up will be null (s was set to -1 by the close routine).
		 * If not, then this is an abnormal close (e.g., a reset)
		 * and clearing out the pointer in the socket structure will
		 * prevent any further operations on what will be a freed
		 * control block. Also wake up anybody waiting on events
		 * related to this block so they will notice it disappearing.
		 */
		if(up != NULLUSOCK){
			up->errcodes[0] = axp->reason;
			up->cb.ax25 = NULLAX25;
		}
		del_ax25(axp);
		break;
	default:	/* Other transitions are ignored */
		break;
	}
	psignal(up,0);	/* In case anybody's waiting */
}
#endif
#ifdef NETROM
/* NET/ROM receive upcall routine */
static void
s_nrcall(cb,cnt)
struct nr4cb *cb;
int cnt;
{
	/* Wake up anybody waiting for data, and let them run */
	struct usock *up;
	up = itop(cb->user);
	psignal(up,1);
	if(up != NULLUSOCK && (up->sigarg & S_INPUT))
	    if(up->sigpid > 0)
		killpg(up->sigpid,SIGPOLL);
	    else
		kill(-up->sigpid,SIGPOLL);
	pwait(NULL);
}
/* NET/ROM transmit upcall routine */
static void
s_ntcall(cb,cnt)
struct nr4cb *cb;
int cnt;
{
	struct usock *up;
	up = itop(cb->user);
	/* Wake up anybody waiting to send data, and let them run */
	psignal(up,1);
	if(up != NULLUSOCK && (up->sigarg & S_OUTPUT))
	    if(up->sigpid > 0)
		killpg(up->sigpid,SIGPOLL);
	    else
		kill(-up->sigpid,SIGPOLL);
	pwait(NULL);
}
/* NET/ROM state change upcall routine */
static void
s_nscall(cb,old,new)
struct nr4cb *cb;
int old,new;
{
	int s,ns;
	struct usock *up,*nup,*oup;
	union sp sp;

	s = cb->user;
	oup = up = itop(s);

	if(new == NR4STDISC && up != NULLUSOCK){
		/* Clean up. If the user has already closed the socket,
		 * then up will be null (s was set to -1 by the close routine).
		 * If not, then this is an abnormal close (e.g., a reset)
		 * and clearing out the pointer in the socket structure will
		 * prevent any further operations on what will be a freed
		 * control block. Also wake up anybody waiting on events
		 * related to this cb so they will notice it disappearing.
		 */
		up->cb.nr4 = NULLNR4CB;
		up->errcodes[0] = cb->dreason;
	}
	if(new == NR4STCON && old == NR4STDISC){
		/* Handle an incoming connection. If this is a server cb,
		 * then we're being handed a "clone" cb and we need to
		 * create a new socket structure for it. In either case,
		 * find out who we're talking to and wake up the guy waiting
		 * for the connection.
		 */
		if(cb->clone){
			/* Clone the socket */
			ns = socket(AF_NETROM,SOCK_SEQPACKET,0);
			nup = itop(ns);
			ASSIGN(*nup,*up);
			cb->user = ns;
			nup->cb.nr4 = cb;
			cb->clone = 0; /* to avoid getting here again */
			/* Allocate new memory for the name areas */
			nup->name = shmalloc(sizeof(struct sockaddr_nr));
			nup->peername = shmalloc(sizeof(struct sockaddr_nr));
			/* Store the new socket # in the old one */
			up->rdysock = ns;
			up = nup;
			s = ns;
		} else {
			/* Allocate space for the peer's name */
			up->peername = shmalloc(sizeof(struct sockaddr_nr));
			/* Store the old socket # in the old socket */
			up->rdysock = s;
		}
		/* Load the addresses. Memory for the name has already
		 * been allocated, either above or in the original bind.
		 */
		sp.p = up->name;
		sp.nr->nr_family = AF_NETROM;
		ASSIGN(sp.nr->nr_addr,up->cb.nr4->local);
		up->namelen = sizeof(struct sockaddr_nr);

		sp.p = up->peername;
		sp.nr->nr_family = AF_NETROM;
		ASSIGN(sp.nr->nr_addr,up->cb.nr4->remote);
		up->peernamelen = sizeof(struct sockaddr_nr);

		/* Wake up the guy accepting it, and let him run */
		psignal(oup,1);
		pwait(NULL);
	}
	/* Ignore all other state transitions */
	psignal(up,0);	/* In case anybody's waiting */
}
#endif

/* Convert a socket index to an internal user socket structure pointer */
static struct usock *
itop(s)
register int s;	/* Socket index */
{
	register struct usock *up;

	s -= USOCKBASE;

	if(s < 0 || s >= Nusock)
		return NULLUSOCK;

	up = &Usock[s];

	if(up->type == NOTUSED)
		return NULLUSOCK;
	return up;
}

/* Restricted clone of socket() used by the upcalls to clone sockets */
int
socket(af,type,protocol)
int af;		/* Address family */
int type;	/* Stream, Datagram or Seqpacket only */
int protocol;	/* Not used here */
{
	register struct usock *up;
	int s,i;

	holdusock();
	for(up=Usock;up < &Usock[Nusock];up++)
		if(up->type == NOTUSED)
			break;

	if(up == &Usock[Nusock]){
		/* None left */
		releaseusock();
		errno = EMFILE;
		return -1;
	}
	s = up - Usock + USOCKBASE;
	errno = 0;
	up->pid[0] = getpid();
	for(i=1; i < sizeof(up->pid)/sizeof(int); ++i)
	    up->pid[i] = 0;
	up->waitcnt = 0;
	up->sigpid = 0;
	up->noblock = 0;
	up->rdysock = -1;
	up->cb.p = NULLCHAR;
	up->peername = up->name = NULLCHAR;
	up->namelen = up->peernamelen = 0;
	memset(up->errcodes,0,sizeof(up->errcodes));
	switch(af){
#ifdef	AX25
	case AF_AX25:
		switch(type){
		case SOCK_STREAM:
			up->type = TYPE_AX25I;
			break;
		case SOCK_DGRAM:
			up->type = TYPE_AX25UI;
			break;
		default:
			errno = ESOCKTNOSUPPORT;
			break;
		}
		break;
#endif
#ifdef NETROM
	case AF_NETROM:
		switch(type){
		case SOCK_SEQPACKET:
			up->type = TYPE_NETROML4;
			break;
		default:
			errno = ESOCKTNOSUPPORT;
			break;
		}
		break;
#endif
	case AF_INET:
		switch(type){
		case SOCK_STREAM:
			up->type = TYPE_TCP;
			break;
		case SOCK_DGRAM:
			up->type = TYPE_UDP;
			break;
		default:
			errno = ESOCKTNOSUPPORT;
			break;
		}
		break;
	default:
		errno = EAFNOSUPPORT;
		break;
	}
	releaseusock();	
	if(errno)
		return -1;

	return s;
}
/* Close down a socket three ways. Type 0 means "no more receives"; this
 * replaces the incoming data upcall with a routine that discards further
 * data. Type 1 means "no more sends", and obviously corresponds to sending
 * a TCP FIN. Type 2 means "no more receives or sends". This I interpret
 * as "abort the connection".
 */
int
shutdown(s,how)
int s;		/* Socket index */
int how;	/* (see above) */
{
	register struct usock *up;

	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	if(up->cb.p == NULLCHAR){
		errno = ENOTCONN;
		return -1;
	}
	switch(up->type){
	case TYPE_TCP:
		switch(how){
		case 0:	/* No more receives -- replace upcall */
			up->cb.tcb->r_upcall = trdiscard;
			break;
		case 1:	/* Send EOF */
			close_tcp(up->cb.tcb);
			break;
		case 2:	/* Blow away TCB */
			reset_tcp(up->cb.tcb);
			up->cb.tcb = NULLTCB;
			break;
		}
		break;
#ifdef	AX25
	case TYPE_AX25UI:
		close_s(s);
		break;
	case TYPE_AX25I:
		switch(how){
		case 0:
		case 1:	/* Attempt regular disconnect */
			disc_ax25(up->cb.ax25);
			break;
		case 2: /* Blow it away */
			reset_ax25(up->cb.ax25);
			up->cb.ax25 = NULLAX25;
			break;
		}
		break;		
#endif
#ifdef	NETROM
	case TYPE_NETROML3:
		close_s(s);
		break;
	case TYPE_NETROML4:
		switch(how){
		case 0:
		case 1:	/* Attempt regular disconnect */
			disc_nr4(up->cb.nr4);
			break;
		case 2: /* Blow it away */
			reset_nr4(up->cb.nr4);
			up->cb.nr4 = NULLNR4CB;
			break;
		}
		break;		
#endif
	case TYPE_RAW:
		close_s(s);
		break;
	default:
		errno = EOPNOTSUPP;
		return -1;
	}
	psignal(up,0);
	return 0;
}
/* Close a socket, freeing it for reuse. Try to do a graceful close on a
 * TCP socket, if possible
 */
int
close_s(s)
int s;		/* Socket index */
{
	register struct usock *up;

	if((up = itop(s)) == NULLUSOCK){
		errno = EBADF;
		return -1;
	}
	switch(up->type){
	case TYPE_TCP:
		if(up->cb.tcb != NULLTCB){	/* In case it's been reset */
			up->cb.tcb->r_upcall = trdiscard;
			/* Tell the CLOSED upcall there's no more socket */
			up->cb.tcb->user = -1;
			close_tcp(up->cb.tcb);
		}
		break;
	case TYPE_UDP:
		if(up->cb.udp != NULLUDP){
			del_udp(up->cb.udp);
		}
		break;
#ifdef	AX25
	case TYPE_AX25I:
		if(up->cb.ax25 != NULLAX25){
			/* Tell the CLOSED upcall there's no more socket */
			up->cb.ax25->user = -1;
			disc_ax25(up->cb.ax25);
		}
		break;
	case TYPE_AX25UI:
		Axui_sock = -1;
		free_q(&Bcq);
		psignal(&Bcq,0);	/* Unblock any reads */
		break;
#endif
#ifdef	NETROM
	case TYPE_NETROML3:
		del_rnr(up->cb.rnr);
		break;
	case TYPE_NETROML4:
		if(up->cb.nr4 != NULLNR4CB){
			/* Tell the CLOSED upcall there's no more socket */
			up->cb.nr4->user = -1;
			disc_nr4(up->cb.nr4);
		}
		break;
#endif
	case TYPE_RAW:
		del_ip(up->cb.rip);
		break;
	default:
		errno = EOPNOTSUPP;
		return -1;
	}
	shfree(up->name);
	shfree(up->peername);

	up->cb.p = NULLCHAR;
	up->name = up->peername = NULLCHAR;
	up->type = NOTUSED;
	psignal(up,0);	/* Wake up anybody doing an accept() or recv() */
	return 0;
}
char *
pinet(s)
struct socket *s;
{
	static char buf[30];

	sprintf(buf,"%s:%u",inet_ntoa(s->address),s->port);
	return buf;
}
